# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
import datetime
from typing import List, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, IPvAnyAddress, field_validator

from catalystwan.endpoints import JSON, APIEndpoints, get, post, put, view
from catalystwan.typed_list import DataSequence
from catalystwan.utils.session_type import ProviderView, SingleTenantView

OnOffMode = Literal["on", "off"]
DataStreamIPTypes = Literal["systemIp", "mgmtIp", "transportIp"]
PasswordPolicies = Literal["disabled", "mediumSecurity", "highSecurity"]
SmartLicensingSettingModes = Literal["on-prem", "offline", "online"]
CRLActions = Literal["disable", "revoke", "quarantine"]


class Organization(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    org: Optional[str] = Field(default=None)
    domain_id: Optional[str] = Field(default=None, serialization_alias="domain-id", validation_alias="domain-id")
    control_connection_up: Optional[bool] = Field(
        default=None, serialization_alias="controlConnectionUp", validation_alias="controlConnectionUp"
    )


class Device(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    domain_ip: Optional[str] = Field(default=None, serialization_alias="domainIp", validation_alias="domainIp")
    port: Optional[str] = Field(default="12346")


class EmailNotificationSettings(BaseModel):
    enabled: Optional[bool] = False


class HardwareRootCA(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    hardware_certificate: Optional[str] = Field(
        default=None, serialization_alias="hardwareCertificate", validation_alias="hardwareCertificate"
    )
    control_connection_up: Optional[bool] = Field(
        default=False, serialization_alias="controlConnectionUp", validation_alias="controlConnectionUp"
    )


class Certificate(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    certificate_signing: str = Field(serialization_alias="certificateSigning", validation_alias="certificateSigning")
    validity_period: Optional[str] = Field(
        default=None, serialization_alias="validityPeriod", validation_alias="validityPeriod"
    )
    retrieve_interval: Optional[str] = Field(
        default=None, serialization_alias="retrieveInterval", validation_alias="retrieveInterval"
    )
    first_name: Optional[str] = Field(default=None, serialization_alias="firstName", validation_alias="firstName")
    last_name: Optional[str] = Field(default=None, serialization_alias="lastName", validation_alias="lastName")
    email: Optional[str] = Field(default=None)


class VEdgeCloud(BaseModel):
    certificateauthority: Optional[str] = None


class Banner(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    mode: Optional[OnOffMode] = "off"
    banner_detail: Optional[str] = Field(serialization_alias="bannerDetail", validation_alias="bannerDetail")


class ProxyHTTPServer(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    proxy: bool
    proxy_ip: str = Field(default="", serialization_alias="proxyIp", validation_alias="proxyIp")
    proxy_port: str = Field(default="", serialization_alias="proxyPort", validation_alias="proxyPort")


class ReverseProxy(BaseModel):
    mode: Optional[OnOffMode] = "off"


class CloudX(BaseModel):
    mode: Optional[OnOffMode] = "off"


class ManageEncryptedPassword(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    manage_type8_password: Optional[bool] = Field(
        default=False, serialization_alias="manageType8Password", validation_alias="manageType8Password"
    )


class CloudServices(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    enabled: Optional[bool] = False
    vanalytics_enabled: Optional[bool] = Field(
        default=False, serialization_alias="vanalyticsEnabled", validation_alias="vanalyticsEnabled"
    )
    vmonitoring_enabled: Optional[bool] = Field(
        default=False, serialization_alias="vmonitoringEnabled", validation_alias="vmonitoringEnabled"
    )
    otp: Optional[str] = None
    cloud_gateway_url: Optional[str] = Field(
        default=None, serialization_alias="cloudGatewayUrl", validation_alias="cloudGatewayUrl"
    )
    vanalytics_enabled_time: Optional[datetime.datetime] = Field(
        default=None, serialization_alias="vanalyticsEnabledTime", validation_alias="vanalyticsEnabledTime"
    )
    vmonitoring_enabled_time: Optional[datetime.datetime] = Field(
        default=None, serialization_alias="vmonitoringEnabledTime", validation_alias="vmonitoringEnabledTime"
    )


class ClientSessionTimeout(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    is_enabled: Optional[bool] = Field(default=False, serialization_alias="isEnabled", validation_alias="isEnabled")
    timeout: Optional[int] = Field(default=None, ge=10, description="timeout in minutes")


class SessionLifeTime(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    session_life_time: int = Field(
        serialization_alias="sessionLifeTime",
        validation_alias="sessionLifeTime",
        ge=30,
        le=10080,
        description="lifetime in minutes",
    )


class ServerSessionTimeout(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    server_session_timeout: int = Field(
        serialization_alias="serverSessionTimeout",
        validation_alias="serverSessionTimeout",
        ge=10,
        le=30,
        description="timeout in minutes",
    )


class MaxSessionsPerUser(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    max_sessions_per_user: int = Field(
        serialization_alias="maxSessionsPerUser", validation_alias="maxSessionsPerUser", ge=1, le=8
    )


class PasswordPolicy(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    password_policy: Union[bool, PasswordPolicies] = Field(
        serialization_alias="passwordPolicy", validation_alias="passwordPolicy"
    )
    password_expiration_time: Optional[int] = Field(
        default=False,
        serialization_alias="passwordExpirationTime",
        validation_alias="passwordExpirationTime",
        ge=1,
        le=90,
        description="timeout in days",
    )


class VManageDataStream(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    enable: Optional[bool] = False
    ip_type: Optional[DataStreamIPTypes] = Field(default=None, serialization_alias="ipType", validation_alias="ipType")
    server_host_name: Union[IPvAnyAddress, DataStreamIPTypes, None] = Field(
        default=None, serialization_alias="serverHostName", validation_alias="serverHostName"
    )
    vpn: Optional[int] = Field(default=None, le=512)


class DataCollectionOnNotification(BaseModel):
    enabled: bool


class SDWANTelemetry(BaseModel):
    enabled: bool


class StatsOperation(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    stats_operation: str = Field(serialization_alias="statsOperation", validation_alias="statsOperation")
    rid: int = Field(serialization_alias="@rid", validation_alias="@rid")
    operation_interval: int = Field(
        serialization_alias="operationInterval",
        validation_alias="operationInterval",
        ge=1,
        description="interval in minutes",
    )
    default_interval: int = Field(
        serialization_alias="defaultInterval",
        validation_alias="defaultInterval",
        ge=1,
        description="interval in minutes",
    )


class MaintenanceWindow(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    enabled: Optional[bool] = False
    message: Optional[str] = ""
    start: Optional[int] = Field(
        default=None, serialization_alias="epochStartTimeInMillis", validation_alias="epochStartTimeInMillis"
    )
    end: Optional[int] = Field(
        default=None, serialization_alias="epochEndTimeInMillis", validation_alias="epochEndTimeInMillis"
    )


class ElasticSearchDBSize(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    index_name: str = Field(serialization_alias="indexName", validation_alias="indexName")
    size_in_gb: int = Field(serialization_alias="sizeInGB", validation_alias="sizeInGB")


class GoogleMapKey(BaseModel):
    key: str


class SoftwareInstallTimeout(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    download_timeout: str = Field(serialization_alias="downloadTimeoutInMin", validation_alias="downloadTimeoutInMin")
    activate_timeout: str = Field(serialization_alias="activateTimeoutInMin", validation_alias="activateTimeoutInMin")
    control_pps: Optional[str] = Field(serialization_alias="controlPps", validation_alias="controlPps")

    @field_validator("download_timeout")
    def check_download_timeout(cls, download_timeout_str: str):
        download_timeout = int(download_timeout_str)
        if download_timeout < 60 or download_timeout > 360:
            raise ValueError("download timeout should be in range 60-360")
        return download_timeout_str

    @field_validator("activate_timeout")
    def check_activate_timeout(cls, activate_timeout_str: str):
        activate_timeout = int(activate_timeout_str)
        if activate_timeout < 30 or activate_timeout > 180:
            raise ValueError("activate timeout should be in range 30-180")
        return activate_timeout_str

    @field_validator("control_pps")
    def check_control_pps(cls, control_pps_str: str):
        control_pps = int(control_pps_str)
        if control_pps < 300 or control_pps > 65535:
            raise ValueError("control pps should be in range 300-65535")
        return control_pps_str


class IPSSignatureSettings(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    is_enabled: Optional[bool] = Field(default=False, serialization_alias="isEnabled", validation_alias="isEnabled")
    username: Optional[str] = None
    update_interval: Optional[int] = Field(
        default=None,
        serialization_alias="updateInterval",
        validation_alias="updateInterval",
        description="interval in minutes",
        ge=1,
        le=1440,
    )


class SmartAccountCredentials(BaseModel):
    username: Optional[str] = None
    password: Optional[str] = None


class PnPConnectSync(BaseModel):
    mode: Optional[OnOffMode] = "off"


class ClaimDevice(BaseModel):
    enabled: bool


class WalkMe(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    walkme: bool
    walkme_analytics: bool = Field(serialization_alias="walkmeAnalytics", validation_alias="walkmeAnalytics")


class SmartLicensingSetting(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    mode: Optional[SmartLicensingSettingModes] = None
    ssm_server_url: Optional[str] = Field(None, serialization_alias="ssmServerUrl", validation_alias="ssmServerUrl")
    ssm_client_id: Optional[str] = Field(None, serialization_alias="ssmClientId", validation_alias="ssmClientId")
    ssm_client_secret: Optional[str] = Field(
        None, serialization_alias="ssmClientSecret", validation_alias="ssmClientSecret"
    )


class StatsCollectionInterval(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    config_name: Literal["statsCollection"] = Field(
        default="statsCollection", serialization_alias="configName", validation_alias="configName"
    )
    operation_interval: int = Field(
        ge=5,
        le=180,
        serialization_alias="operationInterval",
        validation_alias="operationInterval",
        description="collecion interval in minutes",
    )


StatsConfigItem = Union[StatsCollectionInterval, None]  # open for extension for now only one option could be deduced


class StatsConfig(BaseModel):
    config: List[StatsConfigItem]

    @staticmethod
    def from_collection_interval(interval: int) -> "StatsConfig":
        return StatsConfig(config=[StatsCollectionInterval(operation_interval=interval)])


class CRLSettings(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    action: CRLActions
    crl_url: Optional[str] = Field(None, serialization_alias="crlUrl", validation_alias="crlUrl")
    polling_interval: Optional[str] = Field(description="Retrieval interval (1-24 hours)")
    vpn: Optional[str]

    @field_validator("polling_interval")
    def check_polling_interval(cls, polling_interval_str: str):
        polling_interval = int(polling_interval_str)
        if polling_interval < 1 or polling_interval > 24:
            raise ValueError("Polling interval should be in range 1-24")
        return polling_interval_str

    @field_validator("vpn")
    def check_vpn(cls, vpn_str: str):
        vpn = int(vpn_str)
        if vpn < 0 or vpn > 65530:
            raise ValueError("vpn should be in range 0-65530")
        return vpn_str


class ConfigurationSettings(APIEndpoints):
    def create_analytics_data_file(self):
        # POST /settings/configuration/analytics/dca
        ...

    def edit_cert_configuration(self):
        # PUT /settings/configuration/certificate/{settingType}
        ...

    def edit_configuration(self):
        # PUT /settings/configuration/{settingType}
        ...

    def get_cert_configuration(self):
        # GET /settings/configuration/certificate/{settingType}
        ...

    @get("/settings/configuration/{setting_type}")
    def get_configuration_by_setting_type(self, setting_type: str) -> JSON:
        ...

    @get("/settings/configuration/organization", "data")
    def get_organizations(self) -> DataSequence[Organization]:
        ...

    @view({SingleTenantView, ProviderView})
    @get("/settings/configuration/device", "data")
    def get_devices(self) -> DataSequence[Device]:
        ...

    @get("/settings/configuration/emailNotificationSettings", "data")
    def get_email_notification_settings(self) -> DataSequence[EmailNotificationSettings]:
        ...

    @view({SingleTenantView, ProviderView})
    @get("/settings/configuration/hardwarerootca", "data")
    def get_hardware_root_cas(self) -> DataSequence[HardwareRootCA]:
        ...

    @view({SingleTenantView, ProviderView})
    @get("/settings/configuration/certificate", "data")
    def get_certificates(self) -> DataSequence[Certificate]:
        ...

    @view({SingleTenantView, ProviderView})
    @get("/settings/configuration/vedgecloud", "data")
    def get_vedge_cloud(self) -> DataSequence[VEdgeCloud]:
        ...

    @view({SingleTenantView, ProviderView})
    @get("/settings/configuration/crlSetting")
    def get_clr_settings(self) -> DataSequence[CRLSettings]:
        ...

    @get("/settings/configuration/banner", "data")
    def get_banner(self) -> DataSequence[Banner]:
        ...

    @get("/settings/configuration/proxyHttpServer", "data")
    def get_proxy_http_servers(self) -> DataSequence[ProxyHTTPServer]:
        ...

    @get("/settings/configuration/reverseproxy", "data")
    def get_reverse_proxies(self) -> DataSequence[ReverseProxy]:
        ...

    @get("/settings/configuration/cloudx", "data")
    def get_cloudx(self) -> DataSequence[CloudX]:
        ...

    @get("/settings/configuration/manageEncryptedPassword", "data")
    def get_manage_encrypted_password(self) -> DataSequence[ManageEncryptedPassword]:
        ...

    @view({SingleTenantView, ProviderView})
    @get("/settings/configuration/cloudservices", "data")
    def get_cloudservices(self) -> DataSequence[CloudServices]:
        ...

    @get("/settings/configuration/clientSessionTimeout", "data")
    def get_client_session_timeout(self) -> DataSequence[ClientSessionTimeout]:
        ...

    @get("/settings/configuration/sessionLifeTime", "data")
    def get_session_life_time(self) -> DataSequence[SessionLifeTime]:
        ...

    @get("/settings/configuration/serverSessionTimeout", "data")
    def get_server_session_timeout(self) -> DataSequence[ServerSessionTimeout]:
        ...

    @get("/settings/configuration/maxSessionsPerUser", "data")
    def get_max_sessions_per_user(self) -> DataSequence[MaxSessionsPerUser]:
        ...

    @get("/settings/configuration/passwordPolicy", "data")
    def get_password_policy(self) -> DataSequence[PasswordPolicy]:
        ...

    @view({SingleTenantView, ProviderView})
    @get("/settings/configuration/vmanagedatastream", "data")
    def get_vmanage_data_stream(self) -> DataSequence[VManageDataStream]:
        ...

    @view({SingleTenantView, ProviderView})
    @get("/settings/configuration/dataCollectionOnNotification", "data")
    def get_data_collection_on_notification(self) -> DataSequence[DataCollectionOnNotification]:
        ...

    @view({SingleTenantView, ProviderView})
    @get("/settings/configuration/sdWanTelemetry", "data")
    def get_sdwan_telemetry(self) -> DataSequence[SDWANTelemetry]:
        ...

    @view({SingleTenantView, ProviderView})
    @get("/management/statsconfig")
    def get_stats_config(self) -> DataSequence[StatsOperation]:
        ...

    @get("/settings/configuration/spMetadata")
    def get_sp_metadata(self) -> str:
        ...

    @get("/management/elasticsearch/index/size", "indexSize")
    def get_elasticsearch_db_size(self) -> DataSequence[ElasticSearchDBSize]:
        ...

    @get("/settings/configuration/googleMapKey", "data")
    def get_google_map_key(self) -> DataSequence[GoogleMapKey]:
        ...

    @view({SingleTenantView, ProviderView})
    @get("/settings/configuration/maintenanceWindow", "data")
    def get_maintenance_window(self) -> DataSequence[MaintenanceWindow]:
        ...

    @get("/settings/configuration/softwareMaintenance", "data")
    def get_software_install_timeout(self) -> DataSequence[SoftwareInstallTimeout]:
        ...

    @get("/settings/configuration/credentials", "data")
    def get_ips_signature_settings(self) -> DataSequence[IPSSignatureSettings]:
        ...

    @view({SingleTenantView, ProviderView})
    @get("/settings/configuration/smartaccountcredentials", "data")
    def get_smart_account_credentials(self) -> DataSequence[SmartAccountCredentials]:
        ...

    @view({SingleTenantView, ProviderView})
    @get("/settings/configuration/pnpConnectSync", "data")
    def get_pnp_connect_sync(self) -> DataSequence[PnPConnectSync]:
        ...

    @get("/settings/configuration/claimDevice", "data")
    def get_claim_device(self) -> DataSequence[ClaimDevice]:
        ...

    @get("/settings/configuration/walkme", "data")
    def get_walkme(self) -> DataSequence[WalkMe]:
        ...

    @view({SingleTenantView, ProviderView})
    @get("/settings/configuration/smartLicensing", "data")
    def get_smart_licensing_settings(self) -> DataSequence[SmartLicensingSetting]:
        ...

    def new_cert_configuration(self):
        # POST /settings/configuration/certificate/{settingType}
        ...

    def new_configuration(self):
        # POST /settings/configuration/{settingType}
        ...

    @put("/settings/configuration/{setting_type}")
    def edit_configuration_by_setting_type(self, setting_type: str, payload: JSON) -> JSON:
        ...

    @put("/settings/configuration/organization", "data")
    def edit_organizations(self, payload: Organization) -> DataSequence[Organization]:
        ...

    @view({SingleTenantView, ProviderView})
    @put("/settings/configuration/device", "data")
    def edit_devices(self, payload: Device) -> DataSequence[Device]:
        ...

    @put("/settings/configuration/emailNotificationSettings", "data")
    def edit_email_notification_settings(
        self, payload: EmailNotificationSettings
    ) -> DataSequence[EmailNotificationSettings]:
        ...

    @view({SingleTenantView, ProviderView})
    @put("/settings/configuration/hardwarerootca", "data")
    def edit_hardware_root_cas(self, payload: HardwareRootCA) -> DataSequence[HardwareRootCA]:
        ...

    @view({SingleTenantView, ProviderView})
    @put("/settings/configuration/certificate", "data")
    def edit_certificates(self, payload: Certificate) -> DataSequence[Certificate]:
        ...

    @view({SingleTenantView, ProviderView})
    @put("/settings/configuration/vedgecloud", "data")
    def edit_vedge_cloud(self, payload: VEdgeCloud) -> DataSequence[VEdgeCloud]:
        ...

    @put("/settings/configuration/banner", "data")
    def edit_banner(self, payload: Banner) -> DataSequence[Banner]:
        ...

    @put("/settings/configuration/proxyHttpServer", "data")
    def edit_proxy_http_servers(self, payload: ProxyHTTPServer) -> DataSequence[ProxyHTTPServer]:
        ...

    @put("/settings/configuration/reverseproxy", "data")
    def edit_reverse_proxies(self, payload: ReverseProxy) -> DataSequence[ReverseProxy]:
        ...

    @put("/settings/configuration/cloudx", "data")
    def edit_cloudx(self, payload: CloudX) -> DataSequence[CloudX]:
        ...

    @put("/settings/configuration/manageEncryptedPassword", "data")
    def edit_manage_encrypted_password(self, payload: ManageEncryptedPassword) -> DataSequence[ManageEncryptedPassword]:
        ...

    @view({SingleTenantView, ProviderView})
    @put("/settings/configuration/cloudservices", "data")
    def edit_cloudservices(self, payload: CloudServices) -> DataSequence[CloudServices]:
        ...

    @put("/settings/configuration/clientSessionTimeout", "data")
    def edit_client_session_timeout(self, payload: ClientSessionTimeout) -> DataSequence[ClientSessionTimeout]:
        ...

    @put("/settings/configuration/sessionLifeTime", "data")
    def edit_session_life_time(self, payload: SessionLifeTime) -> DataSequence[SessionLifeTime]:
        ...

    @put("/settings/configuration/serverSessionTimeout", "data")
    def edit_server_session_timeout(self, payload: ServerSessionTimeout) -> DataSequence[ServerSessionTimeout]:
        ...

    @put("/settings/configuration/maxSessionsPerUser", "data")
    def edit_max_sessions_per_user(self, payload: MaxSessionsPerUser) -> DataSequence[MaxSessionsPerUser]:
        ...

    @put("/settings/configuration/passwordPolicy", "data")
    def edit_password_policy(self, payload: PasswordPolicy) -> DataSequence[PasswordPolicy]:
        ...

    @view({SingleTenantView, ProviderView})
    @put("/settings/configuration/vmanagedatastream", "data")
    def edit_vmanage_data_stream(self, payload: VManageDataStream) -> DataSequence[VManageDataStream]:
        ...

    @view({SingleTenantView, ProviderView})
    @put("/settings/configuration/dataCollectionOnNotification", "data")
    def edit_data_collection_on_notification(
        self, payload: DataCollectionOnNotification
    ) -> DataSequence[DataCollectionOnNotification]:
        ...

    @view({SingleTenantView, ProviderView})
    @put("/settings/configuration/sdWanTelemetry", "data")
    def edit_sdwan_telemetry(self, payload: SDWANTelemetry) -> DataSequence[SDWANTelemetry]:
        ...

    @view({SingleTenantView, ProviderView})
    @post("/management/statsconfig")
    def edit_stats_config(self, payload: StatsConfig) -> DataSequence[StatsOperation]:
        ...

    @put("/settings/configuration/spMetadata")
    def edit_sp_metadata(self, payload: str) -> str:
        ...

    @put("/management/elasticsearch/index/size", "indexSize")
    def edit_elasticsearch_db_size(self, payload: ElasticSearchDBSize) -> DataSequence[ElasticSearchDBSize]:
        ...

    @put("/settings/configuration/googleMapKey")
    def edit_google_map_key(self, payload: GoogleMapKey) -> DataSequence[GoogleMapKey]:
        ...

    @view({SingleTenantView, ProviderView})
    @put("/settings/configuration/maintenanceWindow")
    def edit_maintenance_window(self, payload: MaintenanceWindow) -> DataSequence[MaintenanceWindow]:
        ...

    @put("/settings/configuration/softwareMaintenance", "data")
    def edit_software_install_timeout(self, payload: SoftwareInstallTimeout) -> DataSequence[SoftwareInstallTimeout]:
        ...

    @put("/settings/configuration/credentials", "data")
    def edit_ips_signature_settings(self, payload: IPSSignatureSettings) -> DataSequence[IPSSignatureSettings]:
        ...

    @view({SingleTenantView, ProviderView})
    @put("/settings/configuration/smartaccountcredentials", "data")
    def edit_smart_account_credentials(self, payload: SmartAccountCredentials) -> DataSequence[SmartAccountCredentials]:
        ...

    @view({SingleTenantView, ProviderView})
    @put("/settings/configuration/pnpConnectSync", "data")
    def edit_pnp_connect_sync(self, payload: PnPConnectSync) -> DataSequence[PnPConnectSync]:
        ...

    @put("/settings/configuration/claimDevice", "data")
    def edit_claim_device(self, payload: ClaimDevice) -> DataSequence[ClaimDevice]:
        ...

    @put("/settings/configuration/walkme", "data")
    def edit_walkme(self, payload: WalkMe) -> DataSequence[WalkMe]:
        ...

    @view({SingleTenantView, ProviderView})
    @put("/settings/configuration/smartLicensing", "data")
    def edit_smart_licensing_settings(self, payload: SmartLicensingSetting) -> DataSequence[SmartLicensingSetting]:
        ...



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from typing import List, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field
from pydantic.functional_validators import BeforeValidator
from typing_extensions import Annotated

from catalystwan.endpoints import APIEndpoints, get, post
from catalystwan.typed_list import DataSequence


def convert_to_list(element: Union[str, List[str]]) -> List[str]:
    return [element] if isinstance(element, str) else element


DeviceType = Literal["vedge", "controller", "vmanage"]

VersionType = Literal["vmanage", "remote"]

PartitionActionType = Literal["removepartition", "defaultpartition", "changepartition"]
LxcActionType = Literal["lxc_activate", "lxc_upgrade", "lxc_delete"]


class ActionId(BaseModel):
    id: str


class GroupId(BaseModel):
    group_id: str = Field(default="all", serialization_alias="groupId", validation_alias="groupId")


class ZTPUpgradeSettings(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    enable_upgrade: bool = Field(serialization_alias="enableUpgrade", validation_alias="enableUpgrade")
    platform_family: str = Field(serialization_alias="platformFamily", validation_alias="platformFamily")
    version_id: str = Field(serialization_alias="versionId", validation_alias="versionId")
    version_name: str = Field(serialization_alias="versionName", validation_alias="versionName")


class PartitionDevice(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    device_id: str = Field(serialization_alias="deviceId", validation_alias="deviceId")
    device_ip: str = Field(serialization_alias="deviceIP", validation_alias="deviceIP")
    version: Union[str, List[str]] = Field(default="")


VersionList = Annotated[Union[str, List[str]], BeforeValidator(convert_to_list)]


class InstallLxcImage(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    network_function_type: str = Field(
        default="app-hosting", serialization_alias="networkFunctionType", validation_alias="networkFunctionType"
    )
    version_name: str = Field(serialization_alias="versionName", validation_alias="versionName")
    version_type_name: str = Field(
        default="UTD-Snort-Feature", serialization_alias="versionTypeName", validation_alias="versionTypeName"
    )


class LxcActivateDevice(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    device_id: str = Field(serialization_alias="deviceId", validation_alias="deviceId")
    device_ip: str = Field(serialization_alias="deviceIP", validation_alias="deviceIP")
    install_images: List[InstallLxcImage] = Field(serialization_alias="installImages", validation_alias="installImages")
    vedge_vpn: str = Field(default="0", serialization_alias="vEdgeVPN", validation_alias="vEdgeVPN")
    version_type: str = Field(default="vmanage", serialization_alias="versionType", validation_alias="versionType")


class LxcUpgradeDevice(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    device_id: str = Field(serialization_alias="deviceId", validation_alias="deviceId")
    device_ip: str = Field(serialization_alias="deviceIP", validation_alias="deviceIP")
    install_images: List[InstallLxcImage] = Field(serialization_alias="installImages", validation_alias="installImages")


class RemovePartitionDevice(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    device_id: str = Field(serialization_alias="deviceId", validation_alias="deviceId")
    device_ip: str = Field(serialization_alias="deviceIP", validation_alias="deviceIP")
    version: VersionList


class PartitionActionPayload(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    action: PartitionActionType
    device_type: str = Field(serialization_alias="deviceType", validation_alias="deviceType")
    devices: List[PartitionDevice]


class LxcInstallInput(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    v_edge_vpn: str = Field(serialization_alias="vEdgeVPN", validation_alias="vEdgeVPN")
    version_type: VersionType = Field(serialization_alias="versionType", validation_alias="versionType")


class LxcImageActivatePayload(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    action: LxcActionType
    device_type: str = Field(serialization_alias="deviceType", validation_alias="deviceType")
    devices: List[LxcActivateDevice]
    input: LxcInstallInput


class LxcImageUpgradePayload(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    action: LxcActionType
    device_type: str = Field(serialization_alias="deviceType", validation_alias="deviceType")
    devices: List[LxcUpgradeDevice]
    input: LxcInstallInput


class LxcImageDeletePayload(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    action: LxcActionType
    device_type: str = Field(serialization_alias="deviceType", validation_alias="deviceType")
    devices: List[LxcUpgradeDevice]


class RemovePartitionActionPayload(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    action: PartitionActionType
    device_type: str = Field(serialization_alias="deviceType", validation_alias="deviceType")
    devices: List[RemovePartitionDevice]


class InstallData(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    family: str
    version: str
    version_id: str = Field(serialization_alias="versionId", validation_alias="versionId")
    remote_server_id: str = Field(serialization_alias="remoteServerId", validation_alias="remoteServerId")


class InstallInput(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    data: Optional[List[InstallData]] = Field(default=None)
    family: Optional[str] = Field(default=None)
    reboot: bool
    sync: bool
    v_edge_vpn: int = Field(serialization_alias="vEdgeVPN", validation_alias="vEdgeVPN")
    v_smart_vpn: int = Field(serialization_alias="vSmartVPN", validation_alias="vSmartVPN")
    version: Optional[str] = Field(default=None)
    version_type: VersionType = Field(serialization_alias="versionType", validation_alias="versionType")


class InstallDevice(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    device_id: str = Field(serialization_alias="deviceId", validation_alias="deviceId")
    device_ip: str = Field(serialization_alias="deviceIP", validation_alias="deviceIP")
    is_nutella_migration: Optional[bool] = Field(
        default=False, serialization_alias="isNutellaMigration", validation_alias="isNutellaMigration"
    )


class InstallActionPayload(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    action: str = Field(default="install")
    device_type: str = Field(serialization_alias="deviceType", validation_alias="deviceType")
    devices: List[InstallDevice]
    input: InstallInput


class InstalledDeviceData(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    available_versions: Optional[List[str]] = Field(
        None, serialization_alias="availableVersions", validation_alias="availableVersions"
    )
    current_partition: Optional[str] = Field(
        None, serialization_alias="current-partition", validation_alias="current-partition"
    )
    default_version: Optional[str] = Field(
        None, serialization_alias="defaultVersion", validation_alias="defaultVersion"
    )
    device_model: Optional[str] = Field(None, serialization_alias="device-model", validation_alias="device-model")
    device_os: Optional[str] = Field(None, serialization_alias="device-os", validation_alias="device-os")
    device_type: Optional[str] = Field(None, serialization_alias="device-type", validation_alias="device-type")
    host_name: Optional[str] = Field(None, serialization_alias="host-name", validation_alias="host-name")
    is_multi_step_upgrade_supported: Optional[bool] = Field(
        None, serialization_alias="isMultiStepUpgradeSupported", validation_alias="isMultiStepUpgradeSupported"
    )
    is_schedule_upgrade_supported: Optional[bool] = Field(
        None, serialization_alias="isScheduleUpgradeSupported", validation_alias="isScheduleUpgradeSupported"
    )
    layout_level: Optional[int] = Field(None, serialization_alias="layoutLevel", validation_alias="layoutLevel")
    local_system_ip: Optional[str] = Field(
        None, serialization_alias="local-system-ip", validation_alias="local-system-ip"
    )
    personality: Optional[str] = Field(None)
    platform: Optional[str] = Field(None)
    platform_family: Optional[str] = Field(
        None, serialization_alias="platformFamily", validation_alias="platformFamily"
    )
    reachability: Optional[str] = Field(None)
    site_id: Optional[str] = Field(None, serialization_alias="site-id", validation_alias="site-id")
    system_ip: Optional[str] = Field(None, serialization_alias="system-ip", validation_alias="system-ip")
    uptime_date: Optional[int] = Field(None, serialization_alias="uptime-date", validation_alias="uptime-date")
    uuid: Optional[str] = Field(None)
    version: Optional[str] = Field(None)


class ConfigurationDeviceActions(APIEndpoints):
    def create_filter_vpn_list(self):
        # GET /device/action/filter/vpn
        ...

    def create_unique_vpn_list(self):
        # POST /device/action/uniquevpnlist
        ...

    def create_vpn_list(self):
        # GET /device/action/vpn
        ...

    def generate_change_partition_info(self):
        # GET /device/action/changepartition
        ...

    def generate_deactivate_info(self):
        # GET /device/action/deactivate
        ...

    def generate_device_action_list(self):
        # GET /device/action/list
        ...

    @get("/device/action/install/devices/{device_type}", "data")
    def get_list_of_installed_devices(
        self, device_type: DeviceType = "controller", params: GroupId = GroupId()
    ) -> DataSequence[InstalledDeviceData]:
        ...

    def generate_install_info(self):
        # GET /device/action/install
        ...

    def generate_reboot_device_list(self):
        # GET /device/action/reboot/devices/{deviceType}
        ...

    def generate_reboot_info(self):
        # GET /device/action/reboot
        ...

    def generate_rediscover_info(self):
        # GET /device/action/rediscover
        ...

    def generate_remove_partition_info(self):
        # GET /device/action/removepartition
        ...

    def generate_security_devices_list(self):
        # GET /device/action/security/devices/{policyType}
        ...

    def get_ztp_upgrade_config(self):
        # GET /device/action/ztp/upgrade
        ...

    @get("/device/action/ztp/upgrade/setting", "data")
    def get_ztp_upgrade_config_setting(self) -> DataSequence[ZTPUpgradeSettings]:
        ...

    def initiate_image_download(self):
        # POST /device/action/image-download
        ...

    def process_amp_api_re_key(self):
        # POST /device/action/security/amp/rekey
        ...

    def process_cancel_task(self):
        # POST /device/action/cancel
        ...

    @post("/device/action/changepartition")
    def process_mark_change_partition(self, payload: PartitionActionPayload) -> ActionId:
        ...

    def process_deactivate_smu(self):
        # POST /device/action/deactivate
        ...

    @post("/device/action/defaultpartition")
    def process_mark_default_partition(self, payload: PartitionActionPayload) -> ActionId:
        ...

    def process_delete_amp_api_key(self):
        # DELETE /device/action/security/amp/apikey/{uuid}
        ...

    @post("/device/action/install")
    def process_install_operation(self, payload: InstallActionPayload) -> ActionId:
        ...

    @post("/device/action/lxcactivate")
    def process_lxc_activate(self, payload: PartitionActionPayload) -> ActionId:
        # POST /device/action/lxcactivate
        ...

    @post("/device/action/lxcdelete")
    def process_lxc_delete(self, payload: PartitionActionPayload) -> ActionId:
        # POST /device/action/lxcdelete
        ...

    def process_lxc_install(self):
        # POST /device/action/lxcinstall
        ...

    def process_lxc_reload(self):
        # POST /device/action/lxcreload
        ...

    def process_lxc_reset(self):
        # POST /device/action/lxcreset
        ...

    @post("/device/action/lxcupgrade")
    def process_lxc_upgrade(self, payload: PartitionActionPayload) -> ActionId:
        # POST /device/action/lxcupgrade
        ...

    def process_reboot(self):
        # POST /device/action/reboot
        ...

    @post("/device/action/removepartition")
    def process_remove_partition(self, payload: RemovePartitionActionPayload) -> ActionId:
        ...

    def process_remove_software_image(self):
        # POST /device/action/image-remove
        ...

    def process_vnf_install(self):
        # POST /device/action/vnfinstall
        ...

    def process_ztp_upgrade_config(self):
        # POST /device/action/ztp/upgrade
        ...

    def process_ztp_upgrade_config_setting(self):
        # POST /device/action/ztp/upgrade/setting
        ...

    def re_discover_all_device(self):
        # POST /device/action/rediscoverall
        ...

    def re_discover_devices(self):
        # POST /device/action/rediscover
        ...

    def test_api_key(self):
        # GET /device/action/security/apikey/{uuid}
        ...

    def test_iox_config(self):
        # GET /device/action/test/ioxconfig/{deviceIP}
        ...

    def trigger_pending_tasks_monitoring(self):
        # GET /device/action/startmonitor
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List

from pydantic import AliasPath, BaseModel, Field, field_validator

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global


class FowardingClassQueueEntry(BaseModel):
    queue: Global[str]

    @field_validator("queue")
    @classmethod
    def check_burst(cls, queue: Global):
        assert 0 <= int(queue.value) <= 7
        return queue


class FowardingClassParcel(_ParcelBase):
    entries: List[FowardingClassQueueEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def add_queue(self, queue: int):
        self.entries.append(FowardingClassQueueEntry(queue=as_global(str(queue))))



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from ipaddress import IPv4Network
from typing import Dict, List, Literal, Set, Tuple, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import Annotated

from catalystwan.models.misc.application_protocols import ApplicationProtocol
from catalystwan.models.policy.policy_definition import (
    AppListEntry,
    DefinitionWithSequencesCommonBase,
    DestinationDataPrefixListEntry,
    DestinationFQDNEntry,
    DestinationGeoLocationEntry,
    DestinationGeoLocationListEntry,
    DestinationIPEntry,
    DestinationPortEntry,
    DestinationPortListEntry,
    LogAction,
    Match,
    PolicyActionType,
    PolicyDefinitionBase,
    PolicyDefinitionSequenceBase,
    ProtocolEntry,
    ProtocolNameEntry,
    ProtocolNameListEntry,
    RuleSetListEntry,
    SourceDataPrefixListEntry,
    SourceFQDNEntry,
    SourceFQDNListEntry,
    SourceGeoLocationEntry,
    SourceGeoLocationListEntry,
    SourceIPEntry,
    SourcePortEntry,
    SourcePortListEntry,
)

ZoneBasedFWPolicySequenceEntry = Annotated[
    Union[
        AppListEntry,
        DestinationDataPrefixListEntry,
        DestinationFQDNEntry,
        DestinationGeoLocationEntry,
        DestinationGeoLocationListEntry,
        DestinationIPEntry,
        DestinationPortEntry,
        DestinationPortListEntry,
        ProtocolEntry,
        ProtocolNameEntry,
        ProtocolNameListEntry,
        RuleSetListEntry,
        SourceDataPrefixListEntry,
        SourceFQDNEntry,
        SourceFQDNListEntry,
        SourceGeoLocationEntry,
        SourceGeoLocationListEntry,
        SourceIPEntry,
        SourcePortEntry,
        SourcePortListEntry,
    ],
    Field(discriminator="field"),
]

ZoneBasedFWPolicySequenceEntryWithRuleSets = Annotated[
    Union[
        AppListEntry,
        RuleSetListEntry,
    ],
    Field(discriminator="field"),
]


class ZoneBasedFWPolicyMatches(Match):
    entries: List[ZoneBasedFWPolicySequenceEntry] = []


class ZoneBasedFWPolicySequenceWithRuleSets(PolicyDefinitionSequenceBase):
    sequence_type: Literal["zoneBasedFW"] = Field(
        default="zoneBasedFW", serialization_alias="sequenceType", validation_alias="sequenceType"
    )
    match: ZoneBasedFWPolicyMatches
    ruleset: bool = True
    actions: List[LogAction] = []
    model_config = ConfigDict(populate_by_name=True)

    def match_rule_set_lists(self, rule_set_ids: Set[UUID]) -> None:
        self._insert_match(RuleSetListEntry.from_rule_set_ids(rule_set_ids))

    def match_app_list(self, app_list_id: UUID) -> None:
        if self.base_action != "inspect":
            raise ValueError("Action must be inspect when Application/Application Family List is selected.")
        self._insert_match(AppListEntry(ref=app_list_id))


class ZoneBasedFWPolicySequence(PolicyDefinitionSequenceBase):
    sequence_type: Literal["zoneBasedFW"] = Field(
        default="zoneBasedFW", serialization_alias="sequenceType", validation_alias="sequenceType"
    )
    match: ZoneBasedFWPolicyMatches
    actions: List[LogAction] = []
    model_config = ConfigDict(populate_by_name=True)

    def match_app_list(self, app_list_id: UUID) -> None:
        if self.base_action != "inspect":
            raise ValueError("Action must be inspect when Application/Application Family List is selected.")
        self._insert_match(AppListEntry(ref=app_list_id))

    def match_destination_data_prefix_list(self, data_prefix_list_id: UUID) -> None:
        self._insert_match(DestinationDataPrefixListEntry(ref=data_prefix_list_id))

    def match_destination_fqdn(self, fqdn: str) -> None:
        self._insert_match(DestinationFQDNEntry(value=fqdn))

    def match_destination_geo_location(self, geo_location: str) -> None:
        self._insert_match(DestinationGeoLocationEntry(value=geo_location))

    def match_destination_geo_location_list(self, geo_location_list_id: UUID) -> None:
        self._insert_match(DestinationGeoLocationListEntry(ref=geo_location_list_id))

    def match_destination_ip(self, networks: List[IPv4Network]) -> None:
        self._insert_match(DestinationIPEntry.from_ipv4_networks(networks))

    def match_destination_ports(self, ports: Set[int] = set(), port_ranges: List[Tuple[int, int]] = []) -> None:
        self._insert_match(DestinationPortEntry.from_port_set_and_ranges(ports, port_ranges))

    def match_destination_port_list(self, port_list_id: UUID) -> None:
        self._insert_match(DestinationPortListEntry(ref=port_list_id))

    def match_protocols(self, protocols: Set[int]) -> None:
        self._insert_match(ProtocolEntry.from_protocol_set(protocols))

    def match_protocol_names(self, names: Set[str], protocol_map: Dict[str, ApplicationProtocol]) -> None:
        app_protocols = []
        for name in names:
            app_protocol = protocol_map.get(name, None)
            if app_protocol is None:
                raise ValueError(f"{name} not found in protocol map keys: {protocol_map.keys()}")
            app_protocols.append(app_protocol)
        self._insert_match(ProtocolNameEntry.from_application_protocols(app_protocols))
        self._insert_match(DestinationPortEntry.from_application_protocols(app_protocols), False)
        self._insert_match(ProtocolEntry.from_application_protocols(app_protocols), False)

    def match_protocol_name_list(self, protocol_name_list_id: UUID) -> None:
        self._insert_match(ProtocolNameListEntry(ref=protocol_name_list_id))

    def match_source_data_prefix_list(self, data_prefix_list_id: UUID) -> None:
        self._insert_match(SourceDataPrefixListEntry(ref=data_prefix_list_id))

    def match_source_fqdn(self, fqdn: str) -> None:
        self._insert_match(SourceFQDNEntry(value=fqdn))

    def match_source_fqdn_list(self, fqdn_list_id: UUID) -> None:
        self._insert_match(SourceFQDNListEntry(ref=fqdn_list_id))

    def match_source_geo_location(self, geo_location: str) -> None:
        self._insert_match(SourceGeoLocationEntry(value=geo_location))

    def match_source_geo_location_list(self, geo_location_list_id: UUID) -> None:
        self._insert_match(SourceGeoLocationListEntry(ref=geo_location_list_id))

    def match_source_ip(self, networks: List[IPv4Network]) -> None:
        self._insert_match(SourceIPEntry.from_ipv4_networks(networks))

    def match_source_port(self, ports: Set[int] = set(), port_ranges: List[Tuple[int, int]] = []) -> None:
        self._insert_match(SourcePortEntry.from_port_set_and_ranges(ports, port_ranges))

    def match_source_port_list(self, port_list_id: UUID) -> None:
        self._insert_match(SourcePortListEntry(ref=port_list_id))


class ZoneBasedFWPolicyEntry(BaseModel):
    source_zone_id: Union[UUID, Literal["self"]] = Field(
        default="self", serialization_alias="sourceZone", validation_alias="sourceZone"
    )
    destination_zone_id: UUID = Field(serialization_alias="destinationZone", validation_alias="destinationZone")
    model_config = ConfigDict(populate_by_name=True)


class ZoneBasedFWPolicyHeader(PolicyDefinitionBase):
    type: Literal["zoneBasedFW"] = "zoneBasedFW"
    mode: str = Field(default="security")
    model_config = ConfigDict(populate_by_name=True)


class ZoneBasedFWPolicyDefinition(DefinitionWithSequencesCommonBase):
    sequences: List[Union[ZoneBasedFWPolicySequence, ZoneBasedFWPolicySequenceWithRuleSets]] = []
    entries: List[ZoneBasedFWPolicyEntry] = []


class ZoneBasedFWPolicy(ZoneBasedFWPolicyHeader):
    type: Literal["zoneBasedFW"] = "zoneBasedFW"
    mode: Literal["security"] = "security"
    definition: ZoneBasedFWPolicyDefinition = ZoneBasedFWPolicyDefinition()

    def add_ipv4_rule(
        self, name: str, base_action: PolicyActionType = "drop", log: bool = False
    ) -> ZoneBasedFWPolicySequence:
        """Adds new IPv4 Rule to Zone Based Firewall Policy

        Args:
            name (str): Rule name
            base_action (BaseAction, optional): Rule base action (drop, pass, inspect) Defaults to BaseAction.DROP.
            log (bool, optional): If true sets log action

        Returns:
            ZoneBasedFWPolicySequence: Rule object for which matches must be added
        """
        sequence = ZoneBasedFWPolicySequence(
            sequence_name=name,
            base_action=base_action,
            sequence_ip_type="ipv4",
            match=ZoneBasedFWPolicyMatches(),
        )
        if log:
            sequence.actions.append(LogAction())
        self.definition.add(sequence)
        return sequence

    def add_ipv4_rule_sets(
        self, name: str, base_action: PolicyActionType = "drop", log: bool = False
    ) -> ZoneBasedFWPolicySequenceWithRuleSets:
        sequence = ZoneBasedFWPolicySequenceWithRuleSets(
            sequence_name=name,
            base_action=base_action,
            sequence_ip_type="ipv4",
            match=ZoneBasedFWPolicyMatches(),
        )
        if log:
            sequence.actions.append(LogAction())
        self.definition.add(sequence)
        return sequence

    def add_zone_pair(self, source_zone_id: UUID, destination_zone_id: UUID) -> None:
        entry = ZoneBasedFWPolicyEntry(
            source_zone_id=source_zone_id,
            destination_zone_id=destination_zone_id,
        )
        self.definition.entries.append(entry)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from typing import Any, List, Literal, Optional, Union, overload
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, model_validator
from typing_extensions import Annotated

from catalystwan.models.policy.policy import (
    AssemblyItemBase,
    PolicyCreationPayload,
    PolicyDefinition,
    PolicyEditPayload,
    PolicyInfo,
)

TrafficDataDirection = Literal[
    "service",
    "tunnel",
    "all",
]

ControlDirection = Literal[
    "in",
    "out",
]


def assert_feature_defintion(definition: Any) -> "CentralizedPolicyDefinition":
    assert isinstance(definition, CentralizedPolicyDefinition)
    return definition


class DataApplicationEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    direction: TrafficDataDirection = "service"
    site_lists: Optional[List[UUID]] = Field(
        default=None, serialization_alias="siteLists", validation_alias="siteLists"
    )
    vpn_lists: List[UUID] = Field(default=[], serialization_alias="vpnLists", validation_alias="vpnLists")
    region_ids: Optional[List[str]] = Field(default=None, serialization_alias="regionIds", validation_alias="regionIds")
    region_lists: Optional[List[UUID]] = Field(
        default=None, serialization_alias="regionLists", validation_alias="regionLists"
    )

    # def apply_site_list(self, site_list_id: UUID):
    #     self.site_lists.append(site_list_id)

    # def apply_vpn_list(self, vpn_list_id: UUID):
    #     self.vpn_lists.append(vpn_list_id)


class ControlApplicationEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    direction: ControlDirection
    site_lists: Optional[List[UUID]] = Field(
        default=None, serialization_alias="siteLists", validation_alias="siteLists"
    )
    region_lists: Optional[List[UUID]] = Field(
        default=None, serialization_alias="regionLists", validation_alias="regionLists"
    )
    region_ids: Optional[List[str]] = Field(default=None, serialization_alias="regionIds", validation_alias="regionIds")


class ControlPolicyItem(AssemblyItemBase):
    type: Literal["control"] = "control"
    entries: List[ControlApplicationEntry] = []

    def assign_to_inbound_sites(self, site_lists: List[UUID]) -> None:
        entry = ControlApplicationEntry(direction="in", site_lists=site_lists)
        self.entries.append(entry)

    def assign_to_outbound_sites(self, site_lists: List[UUID]) -> None:
        entry = ControlApplicationEntry(direction="in", site_lists=site_lists)
        self.entries.append(entry)

    @overload
    def assign_to_inbound_regions(self, *, region_ids: List[int]) -> None:
        ...

    @overload
    def assign_to_inbound_regions(self, *, region_lists: List[UUID]) -> None:
        ...

    def assign_to_inbound_regions(
        self, *, region_ids: Optional[List[int]] = None, region_lists: Optional[List[UUID]] = None
    ) -> None:
        _region_ids = [str(rid) for rid in region_ids] if region_ids else None
        entry = entry = ControlApplicationEntry(direction="in", region_ids=_region_ids, region_lists=region_lists)
        self.entries.append(entry)

    @overload
    def assign_to_outbound_regions(self, *, region_ids: List[int]) -> None:
        ...

    @overload
    def assign_to_outbound_regions(self, *, region_lists: List[UUID]) -> None:
        ...

    def assign_to_outbound_regions(
        self, *, region_ids: Optional[List[int]] = None, region_lists: Optional[List[UUID]] = None
    ) -> None:
        _region_ids = [str(rid) for rid in region_ids] if region_ids else None
        entry = entry = ControlApplicationEntry(direction="out", region_ids=_region_ids, region_lists=region_lists)
        self.entries.append(entry)


class TrafficDataPolicyItem(AssemblyItemBase):
    type: Literal["data"] = "data"
    entries: List[DataApplicationEntry] = []

    @overload
    def assign_to(
        self,
        vpn_lists: List[UUID],
        direction: TrafficDataDirection = "service",
        *,
        site_lists: List[UUID],
    ) -> None:
        ...

    @overload
    def assign_to(
        self,
        vpn_lists: List[UUID],
        direction: TrafficDataDirection = "service",
        *,
        region_lists: List[UUID],
    ) -> None:
        ...

    @overload
    def assign_to(
        self,
        vpn_lists: List[UUID],
        direction: TrafficDataDirection = "service",
        *,
        region_ids: List[int],
    ) -> None:
        ...

    def assign_to(
        self,
        vpn_lists: List[UUID],
        direction: TrafficDataDirection = "service",
        *,
        site_lists: Optional[List[UUID]] = None,
        region_lists: Optional[List[UUID]] = None,
        region_ids: Optional[List[int]] = None,
    ) -> None:
        entry = DataApplicationEntry(
            direction=direction,
            site_lists=site_lists,
            vpn_lists=vpn_lists,
            region_lists=region_lists,
            region_ids=[str(rid) for rid in region_ids] if region_ids else None,
        )
        self.entries.append(entry)


class HubAndSpokePolicyItem(AssemblyItemBase):
    type: Literal["hubAndSpoke"] = "hubAndSpoke"


class MeshPolicyItem(AssemblyItemBase):
    type: Literal["mesh"] = "mesh"


class AppRoutePolicyItem(AssemblyItemBase):
    type: Literal["appRoute"] = "appRoute"


class CFlowDPolicyItem(AssemblyItemBase):
    type: Literal["cflowd"] = "cflowd"


class VpnMembershipGroupPolicyItem(AssemblyItemBase):
    type: Literal["vpnMembershipGroup"] = "vpnMembershipGroup"


AnyAssemblyItem = Annotated[
    Union[
        TrafficDataPolicyItem,
        ControlPolicyItem,
        MeshPolicyItem,
        HubAndSpokePolicyItem,
        AppRoutePolicyItem,
        CFlowDPolicyItem,
        VpnMembershipGroupPolicyItem,
    ],
    Field(discriminator="type"),
]


class CentralizedPolicyDefinition(PolicyDefinition):
    region_role_assembly: List = Field(
        default=[], serialization_alias="regionRoleAssembly", validation_alias="regionRoleAssembly"
    )
    assembly: List[AnyAssemblyItem] = []
    model_config = ConfigDict(populate_by_name=True)


class CentralizedPolicy(PolicyCreationPayload):
    policy_definition: Union[CentralizedPolicyDefinition, str] = Field(
        default=CentralizedPolicyDefinition(),
        serialization_alias="policyDefinition",
        validation_alias="policyDefinition",
    )
    policy_type: Literal["feature", "cli"] = Field(
        default="feature", serialization_alias="policyType", validation_alias="policyType"
    )

    def add_traffic_data_policy(self, traffic_data_policy_id: UUID) -> TrafficDataPolicyItem:
        policy_definition = assert_feature_defintion(self.policy_definition)
        item = TrafficDataPolicyItem(definition_id=traffic_data_policy_id)
        policy_definition.assembly.append(item)
        return item

    def add_control_policy(self, control_policy_id: UUID) -> ControlPolicyItem:
        policy_definition = assert_feature_defintion(self.policy_definition)
        item = ControlPolicyItem(definition_id=control_policy_id)
        policy_definition.assembly.append(item)
        return item

    def add_mesh_policy(self, mesh_policy_id: UUID) -> None:
        policy_definition = assert_feature_defintion(self.policy_definition)
        policy_definition.assembly.append(MeshPolicyItem(definition_id=mesh_policy_id))

    def add_hub_and_spoke_policy(self, hub_and_spoke_policy_id: UUID) -> None:
        policy_definition = assert_feature_defintion(self.policy_definition)
        policy_definition.assembly.append(HubAndSpokePolicyItem(definition_id=hub_and_spoke_policy_id))

    @model_validator(mode="before")
    @classmethod
    def try_parse_policy_definition_string(cls, values):
        # GET /template/policy/vsmart contains string in policyDefinition field
        # while POST /template/policy/vsmart requires a regular object
        # it makes sense to reuse that model for both requests and present parsed data to the user
        # This is only applicable for "feature" policy type
        # when we are trying to deserialize "policyDefinition" field obtained from remote as string
        json_policy_type = values.get("policyType")
        json_policy_definition = values.get("policyDefinition")
        if json_policy_type == "feature":
            if isinstance(json_policy_definition, str):
                values["policyDefinition"] = CentralizedPolicyDefinition.model_validate_json(json_policy_definition)
            else:
                values["policyDefinition"] = CentralizedPolicyDefinition()
        return values

    @model_validator(mode="after")
    def check_definition_content_by_type(self):
        if self.policy_type == "cli":
            assert isinstance(
                self.policy_definition, str
            ), "policy definition must be provided as string for cli policy"
        elif self.policy_type == "feature":
            assert isinstance(
                self.policy_definition, CentralizedPolicyDefinition
            ), "policy definition must be provided as CentralizedPolicyDefinition object for feature policy"
        return self


class CentralizedPolicyEditPayload(PolicyEditPayload, CentralizedPolicy):
    rid: Optional[int] = Field(default=None, serialization_alias="@rid", validation_alias="@rid")


class CentralizedPolicyInfo(PolicyInfo, CentralizedPolicyEditPayload):
    pass



# Copyright 2024 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from typing import Optional

from catalystwan.api.configuration_groups.parcel import _ParcelBase
from catalystwan.endpoints import JSON, APIEndpoints, delete, get, post, put, versions
from catalystwan.models.configuration.feature_profile.common import (
    FeatureProfileCreationPayload,
    FeatureProfileCreationResponse,
    FeatureProfileDetail,
    FeatureProfileEditPayload,
    FeatureProfileInfo,
    GetFeatureProfilesPayload,
    ParcelCreationResponse,
    ParcelId,
    SchemaTypeQuery,
)
from catalystwan.models.configuration.feature_profile.sdwan.management.vpn import ManagementVPN
from catalystwan.models.configuration.feature_profile.sdwan.transport.cellular_controller import CellularController
from catalystwan.typed_list import DataSequence


class TransportFeatureProfile(APIEndpoints):
    @versions(supported_versions=(">=20.13"), raises=False)
    @post("/v1/feature-profile/sdwan/transport")
    def create_transport_feature_profile(
        self, payload: FeatureProfileCreationPayload
    ) -> FeatureProfileCreationResponse:
        ...

    @versions(supported_versions=(">=20.13"), raises=False)
    @get("/v1/feature-profile/sdwan/transport")
    def get_transport_feature_profiles(self, params: GetFeatureProfilesPayload) -> DataSequence[FeatureProfileInfo]:
        ...

    @versions(supported_versions=(">=20.13"), raises=False)
    @get("/v1/feature-profile/sdwan/transport/{profile_id}")
    def get_transport_feature_profile(self, profile_id: str, params: GetFeatureProfilesPayload) -> FeatureProfileDetail:
        ...

    @versions(supported_versions=(">=20.13"), raises=False)
    @put("/v1/feature-profile/sdwan/transport/{profile_id}")
    def edit_transport_feature_profile(
        self, profile_id: str, payload: FeatureProfileEditPayload
    ) -> FeatureProfileCreationResponse:
        ...

    @versions(supported_versions=(">=20.13"), raises=False)
    @delete("/v1/feature-profile/sdwan/transport/{profile_id}")
    def delete_transport_feature_profile(self, profile_id: str) -> None:
        ...

    #
    # ManagementVPN parcel
    #
    @versions(supported_versions=(">=20.13"), raises=False)
    @post("/v1/feature-profile/sdwan/transport/{profile_id}/management/vpn")
    def create_management_vpn_parcel(self, profile_id: str, payload: _ParcelBase) -> ParcelCreationResponse:
        ...

    # @versions(supported_versions=(">=20.13"), raises=False)
    # @get("/v1/feature-profile/sdwan/transport/{profile_id}/management/vpn")
    # def get_management_vpn_parcels(self, profile_id: str) -> ParcelSequence[ManagementVPN]:
    #     ...

    # @versions(supported_versions=(">=20.13"), raises=False)
    # @get("/v1/feature-profile/sdwan/transport/{profile_id}/management/vpn/{parcel_id}")
    # def get_management_vpn_parcel(self, profile_id: str, parcel_id: str) -> Parcel[ManagementVPN]:
    #     ...

    @versions(supported_versions=(">=20.13"), raises=False)
    @put("/v1/feature-profile/sdwan/transport/{profile_id}/management/vpn/{parcel_id}")
    def edit_management_vpn_parcel(
        self, profile_id: str, parcel_id: str, payload: ManagementVPN
    ) -> ParcelCreationResponse:
        ...

    @versions(supported_versions=(">=20.13"), raises=False)
    @delete("/v1/feature-profile/sdwan/transport/{profile_id}/management/vpn/{parcel_id}")
    def delete_management_vpn_parcel(self, profile_id: str, parcel_id: str) -> None:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @get("/v1/feature-profile/sdwan/transport")
    def get_sdwan_transport_feature_profiles(
        self, payload: Optional[GetFeatureProfilesPayload]
    ) -> DataSequence[FeatureProfileInfo]:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @get("/v1/feature-profile/sdwan/transport/cellular-controller/schema", resp_json_key="request")
    def get_sdwan_transport_cellular_controller_parcel_schema(self, params: SchemaTypeQuery) -> JSON:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @post("/v1/feature-profile/sdwan/transport")
    def create_sdwan_transport_feature_profile(
        self, payload: FeatureProfileCreationPayload
    ) -> FeatureProfileCreationResponse:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @delete("/v1/feature-profile/sdwan/transport/{transport_id}")
    def delete_sdwan_transport_feature_profile(self, transport_id: str) -> None:
        ...

    @versions(supported_versions=(">=20.9"), raises=False)
    @post("/v1/feature-profile/sdwan/transport/{transport_id}/cellular-controller")
    def create_cellular_controller_profile_parcel_for_transport(
        self, transport_id: str, payload: CellularController
    ) -> ParcelId:
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List

from pydantic import AliasPath, BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global


class LocalDomainListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    name_server: Global[str] = Field(
        serialization_alias="nameServer", validation_alias="nameServer", description="Ex: cisco.com, *.cisco.com"
    )


class LocalDomainParcel(_ParcelBase):
    entries: List[LocalDomainListEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def from_local_domains(self, domains: List[str]):
        for domain in domains:
            self.entries.append(LocalDomainListEntry(name_server=as_global(domain)))



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable
from catalystwan.models.configuration.feature_profile.sdwan.service.lan.common import (
    IkeCiphersuite,
    IkeGroup,
    IkeMode,
    IpsecCiphersuite,
    PfsGroup,
    TunnelApplication,
)

GreTunnelMode = Literal[
    "ipv4",
    "ipv6",
]


class GreAddress(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    address: Union[Variable, Global[str]]
    mask: Union[Variable, Global[str]]


class TunnelSourceIP(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    tunnel_source: Union[Global[str], Variable] = Field(
        serialization_alias="tunnelSource", validation_alias="tunnelSource"
    )
    tunnel_route_via: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="tunnelRouteVia", validation_alias="tunnelRouteVia", default=None
    )


class TunnelSourceIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    tunnel_source_v6: Union[Global[str], Variable] = Field(
        serialization_alias="tunnelSourceV6", validation_alias="tunnelSourceV6"
    )
    tunnel_route_via: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="tunnelRouteVia", validation_alias="tunnelRouteVia", default=None
    )


class TunnelSourceInterface(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    tunnel_source_interface: Union[Global[str], Variable] = Field(
        serialization_alias="tunnelSourceInterface", validation_alias="tunnelSourceInterface"
    )
    tunnel_route_via: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="tunnelRouteVia", validation_alias="tunnelRouteVia", default=None
    )


class GreSourceIp(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    source_ip: TunnelSourceIP = Field(serialization_alias="sourceIp", validation_alias="sourceIp")


class GreSourceNotLoopback(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    source_not_loopback: TunnelSourceInterface = Field(
        serialization_alias="sourceNotLoopback", validation_alias="sourceNotLoopback"
    )


class GreSourceLoopback(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    source_loopback: TunnelSourceInterface = Field(
        serialization_alias="sourceLoopback", validation_alias="sourceLoopback"
    )


class GreSourceIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    source_ipv6: TunnelSourceIPv6 = Field(serialization_alias="sourceIpv6", validation_alias="sourceIpv6")


class BasicGre(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface_name: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="ifName", validation_alias="ifName", default=None
    )
    description: Optional[Union[Global[str], Variable, Default[None]]] = None
    address: Optional[GreAddress] = None
    ipv6_address: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="ipv6Address", validation_alias="ipv6Address", default=None
    )
    shutdown: Optional[Union[Global[bool], Variable, Default[bool]]] = Default[bool](value=False)
    tunnel_protection: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="tunnelProtection", validation_alias="tunnelProtection", default=Default[bool](value=False)
    )
    tunnel_mode: Optional[Union[Global[GreTunnelMode], Default[GreTunnelMode]]] = Field(
        serialization_alias="tunnelMode",
        validation_alias="tunnelMode",
        default=Default[GreTunnelMode](value="ipv4"),
    )
    tunnel_source_type: Optional[Union[GreSourceIp, GreSourceNotLoopback, GreSourceLoopback, GreSourceIPv6]] = Field(
        serialization_alias="tunnelSourceType", validation_alias="tunnelSourceType", default=None
    )
    tunnel_destination: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="tunnelDestination", validation_alias="tunnelDestination", default=None
    )
    tunnel_destination_v6: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="tunnelDestinationV6", validation_alias="tunnelDestinationV6", default=None
    )
    mtu: Optional[Union[Global[int], Variable, Default[int]]] = Default[int](value=1500)
    mtu_v6: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="mtuV6", validation_alias="mtuV6", default=None
    )
    tcp_mss_adjust: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="tcpMssAdjust", validation_alias="tcpMssAdjust", default=None
    )
    tcp_mss_adjust_v6: Optional[Union[Global[int], Variable, Default[None]]] = Field(
        serialization_alias="tcpMssAdjustV6", validation_alias="tcpMssAdjustV6", default=None
    )
    clear_dont_fragment: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="clearDontFragment",
        validation_alias="clearDontFragment",
        default=Default[bool](value=False),
    )
    dpd_interval: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="dpdInterval", validation_alias="dpdInterval", default=Default[int](value=10)
    )
    dpd_retries: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="dpdRetries", validation_alias="dpdRetries", default=Default[int](value=3)
    )
    ike_version: Optional[Union[Global[int], Default[int]]] = Field(
        serialization_alias="ikeVersion", validation_alias="ikeVersion", default=Default[int](value=1)
    )
    ike_mode: Optional[Union[Global[IkeMode], Variable, Default[IkeMode]]] = Field(
        serialization_alias="ikeMode", validation_alias="ikeMode", default=Default[IkeMode](value="main")
    )
    ike_rekey_interval: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="ikeRekeyInterval", validation_alias="ikeRekeyInterval", default=Default[int](value=14400)
    )
    ike_ciphersuite: Optional[Union[Global[IkeCiphersuite], Variable, Default[IkeCiphersuite]]] = Field(
        serialization_alias="ikeCiphersuite",
        validation_alias="ikeCiphersuite",
        default=Default[IkeCiphersuite](value="aes256-cbc-sha1"),
    )
    ike_group: Optional[Union[Global[IkeGroup], Variable, Default[IkeGroup]]] = Field(
        serialization_alias="ikeGroup", validation_alias="ikeGroup", default=Default[IkeGroup](value="16")
    )
    pre_shared_secret: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="preSharedSecret", validation_alias="preSharedSecret", default=None
    )
    ike_local_id: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="ikeLocalId", validation_alias="ikeLocalId", default=None
    )
    ike_remote_id: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="ikeRemoteId", validation_alias="ikeRemoteId", default=None
    )
    ipsec_rekey_interval: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="ipsecRekeyInterval",
        validation_alias="ipsecRekeyInterval",
        default=Default[int](value=3600),
    )
    ipsec_replay_window: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="ipsecReplayWindow", validation_alias="ipsecReplayWindow", default=Default[int](value=512)
    )
    ipsec_ciphersuite: Optional[Union[Global[IpsecCiphersuite], Variable, Default[IpsecCiphersuite]]] = Field(
        serialization_alias="ipsecCiphersuite",
        validation_alias="ipsecCiphersuite",
        default=Default[IpsecCiphersuite](value="aes256-gcm"),
    )
    perfect_forward_secrecy: Optional[Union[Global[PfsGroup], Variable, Default[PfsGroup]]] = Field(
        serialization_alias="perfectForwardSecrecy",
        validation_alias="perfectForwardSecrecy",
        default=Default[PfsGroup](value="group-16"),
    )


class AdvancedGre(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    application: Optional[Union[Global[TunnelApplication], Variable]] = None


class InterfaceGreData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    basic: BasicGre
    advanced: Optional[AdvancedGre] = None


class InterfaceGreCreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: InterfaceGreData
    metadata: Optional[dict] = None



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global

Action = Literal[
    "reject",
    "accept",
]

Protocol = Literal[
    "IPV4",
    "IPV6",
    "BOTH",
]

Criteria = Literal[
    "or",
    "and",
    "exact",
]

MetricType = Literal[
    "type1",
    "type2",
]

Origin = Literal[
    "egp",
    "igp",
    "incomplete",
]


class StandardCommunityList(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    criteria: Union[Global[Criteria], Default[Criteria]] = Default[Criteria](value="or")
    standard_community_list: List[Global[UUID]] = Field(
        serialization_alias="standardCommunityList", validation_alias="standardCommunityList"
    )


class ExpandedCommunityList(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    expanded_community_list: Global[UUID] = Field(
        serialization_alias="expandedCommunityList", validation_alias="expandedCommunityList"
    )


class RoutePolicyMatch(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    as_path_list: Optional[Global[UUID]] = Field(
        serialization_alias="asPathList", validation_alias="asPathList", default=None
    )
    community_list: Optional[Union[StandardCommunityList, ExpandedCommunityList]] = Field(
        serialization_alias="communityList", validation_alias="communityList", default=None
    )
    ext_community_list: Optional[Global[UUID]] = Field(
        serialization_alias="extCommunityList", validation_alias="extCommunityList", default=None
    )
    bgp_local_preference: Optional[Global[int]] = Field(
        serialization_alias="bgpLocalPreference", validation_alias="bgpLocalPreference", default=None
    )
    metric: Optional[Global[int]] = None
    omp_tag: Optional[Global[int]] = Field(serialization_alias="ompTag", validation_alias="ompTag", default=None)
    ospf_tag: Optional[Global[int]] = Field(serialization_alias="ospfTag", validation_alias="ospfTag", default=None)
    ipv4_address: Optional[Global[UUID]] = Field(
        serialization_alias="ipv4Address", validation_alias="ipv4Address", default=None
    )
    ipv4_nexthop: Optional[Global[UUID]] = Field(
        serialization_alias="ipv4NextHop", validation_alias="ipv4NextHop", default=None
    )
    ipv6_address: Optional[Global[UUID]] = Field(
        serialization_alias="ipv6Address", validation_alias="ipv6Address", default=None
    )
    ipv6_nexthop: Optional[Global[UUID]] = Field(
        serialization_alias="ipv6NextHop", validation_alias="ipv6NextHop", default=None
    )


class AcceptAction(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    enable_accept_action: Default[bool] = Default[bool](value=True)
    set_as_path: Optional[Global[int]] = Field(
        serialization_alias="setAsPath", validation_alias="setAsPath", default=None
    )
    set_community: Optional[Union[Global[bool], Default[bool]]] = Field(
        serialization_alias="setCommunity", validation_alias="setCommunity", default=None
    )
    set_local_preference: Optional[Global[int]] = Field(
        serialization_alias="setLocalPreference", validation_alias="setLocalPreference", default=None
    )
    set_metric: Optional[Global[int]] = Field(
        serialization_alias="setMetric", validation_alias="setMetric", default=None
    )
    set_metric_type: Optional[Global[MetricType]] = Field(
        serialization_alias="setMetricType", validation_alias="setMetricType", default=None
    )
    set_omp_tag: Optional[Global[int]] = Field(
        serialization_alias="setOmpTag", validation_alias="setOmpTag", default=None
    )
    set_origin: Optional[Global[Origin]] = Field(
        serialization_alias="setOrigin", validation_alias="setOrigin", default=None
    )
    set_ospf_tag: Optional[Global[int]] = Field(
        serialization_alias="setOspfTag", validation_alias="setOspfTag", default=None
    )
    set_weight: Optional[Global[int]] = Field(
        serialization_alias="setWeight", validation_alias="setWeight", default=None
    )
    set_ipv4_nexthop: Optional[Global[str]] = Field(
        serialization_alias="setIpv4NextHop", validation_alias="setIpv4NextHop", default=None
    )
    set_ipv6_nexthop: Optional[Global[str]] = Field(
        serialization_alias="setIpv6NextHop", validation_alias="setIpv6NextHop", default=None
    )


class AcceptActions(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    accept: AcceptAction


class RejectActions(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    reject: Default[bool] = Default[bool](value=True)


class RoutePolicySequence(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    sequence_id: Global[int] = Field(serialization_alias="sequenceId", validation_alias="sequenceId")
    sequence_name: Global[str] = Field(serialization_alias="sequenceName", validation_alias="sequenceName")
    base_action: Union[Global[Action], Default[Action]] = Field(
        serialization_alias="baseAction", validation_alias="baseAction", default=Default[Action](value="reject")
    )
    protocol: Union[Global[Protocol], Default[Protocol]] = Default[Protocol](value="IPV4")
    match_entries: Optional[List[RoutePolicyMatch]] = Field(
        serialization_alias="matchEntries", validation_alias="matchEntries", default=None
    )
    actions: Optional[List[Union[AcceptActions, RejectActions]]] = None


class RoutePolicyData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    defautl_action: Union[Global[Action], Default[Action]] = Field(
        serialization_alias="defaultAction",
        validation_alias="defaultAction",
        default=Default[Action](value="reject"),
    )
    sequences: List[RoutePolicySequence] = []


class RoutePolicyCreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: RoutePolicyData
    metadata: Optional[dict] = None



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.models.policy.policy_definition import PolicyDefinitionBase


class Hub(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    site_list: UUID = Field(validation_alias="siteList", serialization_alias="siteList")
    preference: Optional[str] = None
    prefix_lists: List[UUID] = Field(default=[], validation_alias="prefixLists", serialization_alias="prefixLists")
    ipv6_prefix_lists: List[UUID] = Field(
        default=[], validation_alias="ipv6PrefixLists", serialization_alias="ipv6PrefixLists"
    )


class Spoke(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    site_list: UUID = Field(validation_alias="siteList", serialization_alias="siteList")
    hubs: List[Hub]


class HubAndSpokePolicySubDefinition(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    name: str = "My Hub-and-Spoke"
    equal_preference: bool = Field(
        default=True, validation_alias="equalPreference", serialization_alias="equalPreference"
    )
    advertise_tloc: bool = Field(default=False, validation_alias="advertiseTloc", serialization_alias="advertiseTloc")
    tloc_list: Optional[UUID] = Field(default=None, validation_alias="tlocList", serialization_alias="tlocList")
    spokes: List[Spoke] = []


class HubAndSpokePolicyDefinition(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    vpn_list: UUID = Field(validation_alias="vpnList", serialization_alias="vpnList")
    sub_definitions: List[HubAndSpokePolicySubDefinition] = Field(
        default=[], validation_alias="subDefinitions", serialization_alias="subDefinitions"
    )


class HubAndSpokePolicy(PolicyDefinitionBase):
    model_config = ConfigDict(populate_by_name=True)
    type: Literal["hubAndSpoke"] = "hubAndSpoke"
    definition: HubAndSpokePolicyDefinition

    @staticmethod
    def from_vpn_list(name: str, vpn_list: UUID) -> "HubAndSpokePolicy":
        return HubAndSpokePolicy(name=name, definition=HubAndSpokePolicyDefinition(vpn_list=vpn_list))

    def add_hub_and_spoke(
        self,
        name: str,
        hub_site_lists: List[UUID],
        spoke_site_lists: List[UUID],
        advertise_tloc_list: Optional[UUID] = None,
    ) -> HubAndSpokePolicySubDefinition:
        # supports basic configuration with equal preference
        hubs = [Hub(site_list=hub_site_id) for hub_site_id in hub_site_lists]
        spokes = [Spoke(site_list=spoke_site_id, hubs=hubs.copy()) for spoke_site_id in spoke_site_lists]
        sub_definition = HubAndSpokePolicySubDefinition(
            name=name,
            equal_preference=True,
            advertise_tloc=(advertise_tloc_list is not None),
            tloc_list=advertise_tloc_list,
            spokes=spokes,
        )
        self.definition.sub_definitions.append(sub_definition)
        return sub_definition



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import unittest
from typing import Any, List, Optional
from unittest.mock import patch

from attr import define, field  # type: ignore
from parameterized import parameterized  # type: ignore
from pydantic import BaseModel, Field, ValidationError
from requests.exceptions import JSONDecodeError

from catalystwan.dataclasses import DataclassBase
from catalystwan.response import ManagerErrorInfo, ManagerResponse
from catalystwan.typed_list import DataSequence


@define
class ParsedDataTypeAttrs(DataclassBase):
    key1: str
    key2: int
    key3: Optional[float] = field(default=None)


class ParsedDataTypePydanticV2(BaseModel):
    key1: str
    key2: int
    key3: Optional[float] = Field(default=None)


class DataForValidateTest(BaseModel):
    important: str
    not_important: ParsedDataTypePydanticV2


PARSE_DATASEQ_TEST_DATA: List = [
    (True, None, 0, "data"),
    (True, "string", 0, "data"),
    (True, {}, 0, "data"),
    (True, {"key1": "string", "key2": 12, "key3": 4.5}, 0, "data"),
    (True, 15, 0, "data"),
    (True, {"data": None}, 0, "data"),
    (True, {"data": {}}, 0, "data"),
    (False, {"data": []}, 0, "data"),
    (True, {"data": "something"}, 0, "data"),
    (False, {"data": {"key1": "string", "key2": 12, "key3": None}}, 1, "data"),
    (True, {"data": {"key1": "string", "key2": 12, "key3": None}}, 1, "other"),
    (False, {"other": {"key1": "string", "key2": 12, "key3": None}}, 1, "other"),
    (False, {"data": [{"key1": "string", "key2": 66}, {"key1": "required", "key2": 18, "key3": 0.1}]}, 2, "data"),
    (True, {"headers": {"key1": 1, "key2": "two"}}, 0, "data"),
    (True, {"error": {"message": "Error happened!", "details": "error details", "code": "ABC123"}}, 0, "data"),
]


PARSE_DATAOBJ_TEST_DATA: List = [
    (True, None, "data"),
    (True, "string", "data"),
    (True, {}, "data"),
    (True, {"data": "something"}, "data"),
    (False, {"data": {"key1": "string", "key2": 12, "key3": None}}, "data"),
    (True, {"data": {"key1": "string", "key2": 12, "key3": None}}, "other"),
    (False, {"other": {"key1": "string", "key2": 12, "key3": 55.13}}, "other"),
    (True, {"data": [{"key1": "string", "key2": 66}, {"key1": "required", "key2": 18, "key3": 0.1}]}, "data"),
]

VALIDATE_DATASEQ_TEST_DATA = [
    {"important": "correct1", "not_important": {"invalid-key1": "string", "key2": "not int"}},
    {"important": "correct2", "not_important": {}},
]


class TestResponse(unittest.TestCase):
    @patch("requests.Response")
    def setUp(self, response_mock) -> None:
        self.response_mock = response_mock
        self.response_mock.headers = {"set-cookie": ""}
        self.response_mock.cookies = {}
        self.response_mock.status_code = 200

    @parameterized.expand(PARSE_DATASEQ_TEST_DATA)
    def test_dataseq_attrs(self, raises: bool, json: Any, expected_len: int, sourcekey: str):
        self.response_mock.json.return_value = json
        vmng_response = ManagerResponse(self.response_mock)
        if not raises:
            data_sequence = vmng_response.dataseq(ParsedDataTypeAttrs, sourcekey)
            assert isinstance(data_sequence, DataSequence)
            assert len(data_sequence) == expected_len
        else:
            with self.assertRaises(Exception):
                vmng_response.dataseq(ParsedDataTypeAttrs, sourcekey)

    @parameterized.expand(PARSE_DATASEQ_TEST_DATA)
    def test_dataseq_pydantic(self, raises: bool, json: Any, expected_len: int, sourcekey: str):
        self.response_mock.json.return_value = json
        vmng_response = ManagerResponse(self.response_mock)
        if not raises:
            data_sequence = vmng_response.dataseq(ParsedDataTypePydanticV2, sourcekey)
            assert isinstance(data_sequence, DataSequence)
            assert len(data_sequence) == expected_len
        else:
            with self.assertRaises(Exception):
                vmng_response.dataseq(ParsedDataTypePydanticV2, sourcekey)

    @parameterized.expand(PARSE_DATAOBJ_TEST_DATA)
    def test_dataobj_attrs(self, raises: bool, json: Any, sourcekey: str):
        self.response_mock.json.return_value = json
        vmng_response = ManagerResponse(self.response_mock)
        if not raises:
            data_object = vmng_response.dataobj(ParsedDataTypeAttrs, sourcekey)
            assert isinstance(data_object, ParsedDataTypeAttrs)
        else:
            with self.assertRaises(Exception):
                vmng_response.dataobj(ParsedDataTypeAttrs, sourcekey)

    @parameterized.expand(PARSE_DATAOBJ_TEST_DATA)
    def test_dataobj_pydantic(self, raises: bool, json: Any, sourcekey: str):
        self.response_mock.json.return_value = json
        vmng_response = ManagerResponse(self.response_mock)
        if not raises:
            data_object = vmng_response.dataobj(ParsedDataTypePydanticV2, sourcekey)
            assert isinstance(data_object, ParsedDataTypePydanticV2)
        else:
            with self.assertRaises(Exception):
                vmng_response.dataobj(ParsedDataTypePydanticV2, sourcekey)

    @parameterized.expand(
        [
            (True, "string"),
            (True, {}),
            (True, {"key1": "string", "key2": 12, "key3": 4.5}),
            (True, 15),
            (True, {"data": None}),
            (True, {"data": {}}),
            (True, {"data": []}),
            (True, {"data": "something"}),
            (True, {"data": {"key1": "string", "key2": 12, "key3": -0.7}}),
            (True, {"data": [{"key1": "string", "key2": 66}, {"key1": "required", "key2": 18, "key3": 0.1}]}),
            (True, {"headers": {"key1": 1, "key2": "two"}}),
            (False, {"error": {"message": "Error happened!", "details": "error details", "code": "ABC123"}}),
        ]
    )
    def test_get_error(self, empty_error: bool, json: Any):
        self.response_mock.json.return_value = json
        vmng_response = ManagerResponse(self.response_mock)
        error_info = vmng_response.get_error_info()
        assert isinstance(error_info, ManagerErrorInfo)
        if empty_error:
            assert error_info.message is None
            assert error_info.details is None
            assert error_info.code is None

    def test_dataobj_optional_validate(self):
        # Arrange
        self.response_mock.json.return_value = VALIDATE_DATASEQ_TEST_DATA[0]
        vmng_response = ManagerResponse(self.response_mock)
        # Act
        data = vmng_response.dataobj(DataForValidateTest, sourcekey=None, validate=False)
        # Assert
        assert isinstance(data, DataForValidateTest)
        assert data.important == VALIDATE_DATASEQ_TEST_DATA[0]["important"]
        with self.assertRaises(ValidationError):
            vmng_response.dataobj(DataForValidateTest, sourcekey=None, validate=True)

    def test_dataseq_optional_validate(self):
        self.response_mock.json.return_value = VALIDATE_DATASEQ_TEST_DATA
        vmng_response = ManagerResponse(self.response_mock)
        # Act
        dataseq = vmng_response.dataseq(DataForValidateTest, sourcekey=None, validate=False)
        # Assert
        assert isinstance(dataseq, DataSequence)
        for i, data in enumerate(dataseq):
            assert isinstance(data, DataForValidateTest)
            assert data.important == VALIDATE_DATASEQ_TEST_DATA[i]["important"]
        with self.assertRaises(ValidationError):
            vmng_response.dataseq(DataForValidateTest, sourcekey=None, validate=True)

    def test_dataseq_with_misisng_data(self):
        self.response_mock.json.side_effect = JSONDecodeError("test", "test", 1)
        vmng_response = ManagerResponse(self.response_mock)
        # Act
        dataseq = vmng_response.dataseq(DataForValidateTest, sourcekey="data", validate=True)
        # Assert
        assert isinstance(dataseq, DataSequence)
        assert len(dataseq) == 0



# Copyright 2022 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import logging
from enum import Enum
from functools import cached_property
from pathlib import Path
from time import monotonic, sleep
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Dict, List, Optional, Union
from urllib.parse import urljoin, urlparse, urlunparse

from packaging.version import Version  # type: ignore
from requests import PreparedRequest, Request, Response, Session, get, head
from requests.auth import AuthBase
from requests.exceptions import ConnectionError, HTTPError, RequestException

from catalystwan import USER_AGENT
from catalystwan.endpoints import APIEndpointClient
from catalystwan.endpoints.client import AboutInfo, ServerInfo
from catalystwan.exceptions import (
    DefaultPasswordError,
    ManagerHTTPError,
    ManagerReadyTimeout,
    ManagerRequestException,
    SessionNotCreatedError,
    TenantSubdomainNotFound,
)
from catalystwan.models.tenant import Tenant
from catalystwan.response import ManagerResponse, response_history_debug
from catalystwan.utils.session_type import SessionType
from catalystwan.version import NullVersion, parse_api_version
from catalystwan.vmanage_auth import vManageAuth

JSON = Union[Dict[str, "JSON"], List["JSON"], str, int, float, bool, None]

if TYPE_CHECKING:
    from catalystwan.api.api_container import APIContainer
    from catalystwan.endpoints.endpoints_container import APIEndpointContainter


class UserMode(str, Enum):
    PROVIDER = "provider"
    TENANT = "tenant"


class ViewMode(str, Enum):
    PROVIDER = "provider"
    TENANT = "tenant"


class TenancyMode(str, Enum):
    SINGLE_TENANT = "SingleTenant"
    MULTI_TENANT = "MultiTenant"


class ManagerSessionState(Enum):
    # there are some similiarities to state-machine but flow is only in one direction
    # and does not depend on external inputs
    RESTART_IMMINENT = 0
    WAIT_SERVER_READY_AFTER_RESTART = 1
    LOGIN = 2
    OPERATIVE = 3


def determine_session_type(
    tenancy_mode: Optional[str], user_mode: Optional[str], view_mode: Optional[str]
) -> SessionType:
    modes_map = {
        (TenancyMode.SINGLE_TENANT, UserMode.TENANT, ViewMode.TENANT): SessionType.SINGLE_TENANT,
        (TenancyMode.MULTI_TENANT, UserMode.PROVIDER, ViewMode.PROVIDER): SessionType.PROVIDER,
        (TenancyMode.MULTI_TENANT, UserMode.PROVIDER, ViewMode.TENANT): SessionType.PROVIDER_AS_TENANT,
        (TenancyMode.MULTI_TENANT, UserMode.TENANT, ViewMode.TENANT): SessionType.TENANT,
    }
    try:
        return modes_map.get(
            (TenancyMode(tenancy_mode), UserMode(user_mode), ViewMode(view_mode)), SessionType.NOT_DEFINED
        )
    except ValueError:
        return SessionType.NOT_DEFINED


def create_manager_session(
    url: str,
    username: str,
    password: str,
    port: Optional[int] = None,
    subdomain: Optional[str] = None,
    logger: Optional[logging.Logger] = None,
) -> ManagerSession:
    """Factory method that creates session object and performs login according to parameters

    Args:
        url (str): IP address or domain name
        username (str): username
        password (str): password
        port (int): port
        subdomain: subdomain specifying to which view switch when creating provider as a tenant session,
            works only on provider user mode
        logger: override default module logger

    Returns:
        ManagerSession: logged-in and operative session to perform tasks on SDWAN Manager.
    """
    session = ManagerSession(url=url, username=username, password=password, port=port, subdomain=subdomain)

    if logger:
        session.logger = logger

    session.state = ManagerSessionState.LOGIN
    session.on_session_create_hook()
    return session


class ManagerResponseAdapter(Session):
    def request(self, method, url, *args, **kwargs) -> ManagerResponse:
        return ManagerResponse(super().request(method, url, *args, **kwargs))

    def get(self, url, *args, **kwargs) -> ManagerResponse:
        return ManagerResponse(super().get(url, *args, **kwargs))

    def post(self, url, *args, **kwargs) -> ManagerResponse:
        return ManagerResponse(super().post(url, *args, **kwargs))

    def put(self, url, *args, **kwargs) -> ManagerResponse:
        return ManagerResponse(super().put(url, *args, **kwargs))

    def delete(self, url, *args, **kwargs) -> ManagerResponse:
        return ManagerResponse(super().delete(url, *args, **kwargs))


class ManagerSession(ManagerResponseAdapter, APIEndpointClient):
    """Base class for API sessions for vManage client.

    Defines methods and handles session connectivity available for provider, provider as tenant, and tenant.

    Args:
        url: IP address or domain name, i.e. '10.0.1.200' or 'example.com'
        port: port
        username: username
        password: password

    Attributes:
        enable_relogin (bool): defaults to True, in case that session is not properly logged-in, session will try to
            relogin and try the same request again
    """

    on_session_create_hook: ClassVar[Callable[[ManagerSession], Any]] = lambda *args: None

    def __init__(
        self,
        url: str,
        username: str,
        password: str,
        verify: bool = False,
        port: Optional[int] = None,
        subdomain: Optional[str] = None,
        auth: Optional[AuthBase] = None,
        validate_responses: bool = True,
    ):
        self.url = url
        self.port = port
        self.base_url = self.__create_base_url()
        self.username = username
        self.password = password
        self.subdomain = subdomain
        self._session_type = SessionType.NOT_DEFINED
        self.server_name: Optional[str] = None
        self.logger = logging.getLogger(__name__)
        self.enable_relogin: bool = True
        self.response_trace: Callable[
            [Optional[Response], Union[Request, PreparedRequest, None]], str
        ] = response_history_debug
        super(ManagerSession, self).__init__()
        self.headers.update({"User-Agent": USER_AGENT})
        self.__prepare_session(verify, auth)
        self._platform_version: str = ""
        self._api_version: Version
        self._state: ManagerSessionState = ManagerSessionState.OPERATIVE
        self.restart_timeout: int = 1200
        self.polling_requests_timeout: int = 10
        self._validate_responses = validate_responses

    @cached_property
    def api(self) -> APIContainer:
        from catalystwan.api.api_container import APIContainer

        self._api = APIContainer(self)
        return self._api

    @cached_property
    def endpoints(self) -> APIEndpointContainter:
        from catalystwan.endpoints.endpoints_container import APIEndpointContainter

        self._endpoints = APIEndpointContainter(self)
        return self._endpoints

    @property
    def state(self) -> ManagerSessionState:
        return self._state

    @state.setter
    def state(self, state: ManagerSessionState) -> None:
        """Resets the session to given state and manages transition to desired OPERATIONAL state"""
        self._state = state
        self.logger.debug(f"Session entered state: {self.state.name}")

        if state == ManagerSessionState.OPERATIVE:
            # this is desired state, nothing to be done
            return
        elif state == ManagerSessionState.RESTART_IMMINENT:
            # in this state we process requests normally
            # but when ConnectionError is caught we enter WAIT_SERVER_READY_AFTER_RESTART
            # state change is achieved with cooperation with request method
            return
        elif state == ManagerSessionState.WAIT_SERVER_READY_AFTER_RESTART:
            self.wait_server_ready(self.restart_timeout)
            self.state = ManagerSessionState.LOGIN
        elif state == ManagerSessionState.LOGIN:
            self.login()
            self.state = ManagerSessionState.OPERATIVE
        return

    def restart_imminent(self, restart_timeout_override: Optional[int] = None):
        """Notify session that restart is imminent.
        ConnectionError and status code 503 will cause session to wait for connectivity and perform login again

        Args:
            restart_timeout_override (Optional[int], optional): override session property which controls restart timeout
        """
        if restart_timeout_override is not None:
            self.restart_timeout = restart_timeout_override
        self.state = ManagerSessionState.RESTART_IMMINENT

    def login(self) -> ManagerSession:
        """Performs login to SDWAN Manager and fetches important server info to instance variables

        Raises:
            SessionNotCreatedError: indicates session configuration is not consistent

        Returns:
            ManagerSession: (self)
        """

        self.cookies.clear_session_cookies()
        self.auth = vManageAuth(self.base_url, self.username, self.password, verify=False)
        self.auth.logger = self.logger

        if self.subdomain:
            tenant_id = self.get_tenant_id()
            vsession_id = self.get_virtual_session_id(tenant_id)
            self.headers.update({"VSessionId": vsession_id})
        try:
            server_info = self.server()
        except DefaultPasswordError:
            server_info = ServerInfo.model_construct(**{})

        self.server_name = server_info.server

        tenancy_mode = server_info.tenancy_mode
        user_mode = server_info.user_mode
        view_mode = server_info.view_mode

        self._session_type = determine_session_type(tenancy_mode, user_mode, view_mode)
        if user_mode is UserMode.TENANT and self.subdomain:
            raise SessionNotCreatedError(
                f"Session not created. Subdomain {self.subdomain} passed to tenant session, "
                "cannot switch to tenant from tenant user mode."
            )
        elif self._session_type is SessionType.NOT_DEFINED:
            self.logger.warning(
                "Cannot determine session type for "
                f"tenancy-mode: {tenancy_mode}, user-mode: {user_mode}, view-mode: {view_mode}"
            )

        self.logger.info(
            f"Logged to vManage({self.platform_version}) as {self.username}. The session type is {self.session_type}"
        )
        if jsessionid := self.auth.set_cookie.get("JSESSIONID"):
            self.cookies.set("JSESSIONID", jsessionid)
        return self

    def wait_server_ready(self, timeout: int, poll_period: int = 10) -> None:
        """Waits until server is ready for API requests with given timeout in seconds"""

        begin = monotonic()
        self.logger.info(f"Waiting for server ready with timeout {timeout} seconds.")

        def elapsed() -> float:
            return monotonic() - begin

        # wait for http available
        while elapsed() < timeout:
            available = False
            try:
                resp = head(
                    self.base_url,
                    timeout=self.polling_requests_timeout,
                    verify=False,
                    headers={"User-Agent": USER_AGENT},
                )
                self.logger.debug(self.response_trace(resp, None))
                if resp.status_code != 503:
                    available = True
            except ConnectionError as error:
                self.logger.debug(self.response_trace(error.response, error.request))
            if not available:
                sleep(poll_period)
                continue
            break

        # wait server ready flag
        server_ready_url = self.get_full_url("/dataservice/client/server/ready")
        while elapsed() < timeout:
            try:
                resp = get(
                    server_ready_url,
                    timeout=self.polling_requests_timeout,
                    verify=False,
                    headers={"User-Agent": USER_AGENT},
                )
                self.logger.debug(self.response_trace(resp, None))
                if resp.status_code == 200:
                    if resp.json().get("isServerReady") is True:
                        self.logger.debug(f"Waiting for server ready took: {elapsed()} seconds.")
                        return
                sleep(poll_period)
                continue
            except RequestException as exception:
                self.logger.debug(self.response_trace(exception.response, exception.request))
                raise ManagerRequestException(*exception.args)

        raise ManagerReadyTimeout(f"Waiting for server ready took longer than {timeout} seconds.")

    def request(self, method, url, *args, **kwargs) -> ManagerResponse:
        full_url = self.get_full_url(url)
        try:
            response = super(ManagerSession, self).request(method, full_url, *args, **kwargs)
            self.logger.debug(self.response_trace(response, None))
            if self.state == ManagerSessionState.RESTART_IMMINENT and response.status_code == 503:
                self.state = ManagerSessionState.WAIT_SERVER_READY_AFTER_RESTART
        except RequestException as exception:
            self.logger.debug(self.response_trace(exception.response, exception.request))
            if self.state == ManagerSessionState.RESTART_IMMINENT and isinstance(exception, ConnectionError):
                self.state = ManagerSessionState.WAIT_SERVER_READY_AFTER_RESTART
                return self.request(method, url, *args, **kwargs)
            self.logger.debug(exception)
            raise ManagerRequestException(*exception.args, request=exception.request, response=exception.response)

        if self.enable_relogin and response.jsessionid_expired and self.state == ManagerSessionState.OPERATIVE:
            self.logger.warning("Logging to session. Reason: expired JSESSIONID detected in response headers")
            self.state = ManagerSessionState.LOGIN
            return self.request(method, url, *args, **kwargs)

        if response.request.url and "passwordReset.html" in response.request.url:
            raise DefaultPasswordError("Password must be changed to use this session.")

        try:
            response.raise_for_status()
        except HTTPError as error:
            self.logger.debug(error)
            error_info = response.get_error_info()
            raise ManagerHTTPError(*error.args, error_info=error_info, request=error.request, response=error.response)
        return response

    def get_full_url(self, url_path: str) -> str:
        """Returns base API url plus given url path."""
        return urljoin(self.base_url, url_path)

    def __create_base_url(self) -> str:
        """Creates base url based on ip address or domain and port if provided.

        Returns:
            str: Base url shared for every request.
        """
        url = urlparse(self.url)
        netloc: str = url.netloc or url.path
        scheme: str = url.scheme or "https"
        base_url = urlunparse((scheme, netloc, "", None, None, None))
        if self.port:
            return f"{base_url}:{self.port}"
        return base_url

    def about(self) -> AboutInfo:
        return self.endpoints.client.about()

    def server(self) -> ServerInfo:
        server_info = self.endpoints.client.server()
        self.platform_version = server_info.platform_version
        return server_info

    def get_data(self, url: str) -> Any:
        return self.get_json(url)["data"]

    def get_json(self, url: str) -> Any:
        response = self.get(url)
        return response.json()

    def get_file(self, url: str, filename: Path) -> Response:
        """Get a file using session get.

        Args:
            url: dataservice api.
            filename: Filename to write download file to.

        Returns:
            http response.

        Example usage:
            response = self.session.get_file(url, filename)

        """
        with self.get(url) as response:
            with open(filename, "wb") as file:
                file.write(response.content)
        return response

    def get_tenant_id(self) -> str:
        """Gets tenant UUID for its subdomain.

        Returns:
            Tenant UUID.
        """
        tenants = self.get("dataservice/tenant").dataseq(Tenant)
        tenant = tenants.filter(subdomain=self.subdomain).single_or_default()

        if not tenant or not tenant.tenant_id:
            raise TenantSubdomainNotFound(f"Tenant ID for sub-domain: {self.subdomain} not found")

        return tenant.tenant_id

    def get_virtual_session_id(self, tenant_id: str) -> str:
        """Get VSessionId for a specific tenant

        Note: In a multitenant vManage system, this API is only available in the Provider view.

        Args:
            tenant_id: provider or tenant UUID
        Returns:
            Virtual session token
        """
        url_path = f"/dataservice/tenant/{tenant_id}/vsessionid"
        response = self.post(url_path)
        return response.json()["VSessionId"]

    def logout(self) -> Optional[ManagerResponse]:
        response = None
        if isinstance((version := self.api_version), NullVersion):
            self.logger.warning("Cannot perform logout operation without known api_version.")
            return response
        else:
            # disable automatic relogin before performing logout request
            _relogin = self.enable_relogin
            try:
                self.enable_relogin = False
                if version >= Version("20.12"):
                    response = self.post("/logout")
                else:
                    response = self.get("/logout")
            finally:
                # restore original setting after performing logout request
                self.enable_relogin = _relogin
        return response

    def close(self) -> None:
        """Closes the ManagerSession.

        This method is overrided from requests.Session.
        Firstly it cleans up any resources associated with vManage.
        Then it closes all adapters and as such the session.

        Note: It is generally recommended to use the session as a context manager
        using the `with` statement, which ensures that the session is properly
        closed and resources are cleaned up even in case of exceptions.
        """
        self.logout()
        super().close()

    def __prepare_session(self, verify: bool, auth: Optional[AuthBase]) -> None:
        self.auth = auth
        self.verify = verify

    @property
    def session_type(self) -> SessionType:
        return self._session_type

    @property
    def platform_version(self) -> str:
        return self._platform_version

    @platform_version.setter
    def platform_version(self, version: str):
        self._platform_version = version
        self._api_version = parse_api_version(version)

    @property
    def api_version(self) -> Version:
        return self._api_version

    @property
    def validate_responses(self) -> bool:
        return self._validate_responses

    @validate_responses.setter
    def validate_responses(self, value: bool):
        self._validate_responses = value

    def __str__(self) -> str:
        return f"{self.username}@{self.base_url}"

    def __repr__(self):
        return (
            f"{self.__class__.__name__}('{self.url}', '{self.username}', '{self.password}', port={self.port}, "
            f"subdomain='{self.subdomain}')"
        )

    def __eq__(self, other):
        if isinstance(other, ManagerSession):
            comparison_list = [
                self.url == other.url,
                self.username == other.username,
                self.password == other.password,
                self.port == other.port,
                str(self.subdomain) == str(other.subdomain),
            ]
            return True if all(comparison_list) else False
        return False



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch
from uuid import uuid4

from catalystwan.api.tenant_backup_restore_api import TenantBackupRestoreAPI


class TestTenantBackupRestoreAPI(unittest.TestCase):
    def setUp(self):
        self.processId = str(uuid4())
        self.download_file_content = "Downloaded file content"
        self.full_name = Path("/dataservice")
        self.full_name = (
            self.full_name
            / "tenantbackup"
            / "download"
            / "3cf632d2-5a62-45cb-af50-50c5ef1b1202"
            / "bkup_3cf632d2-5a62-45cb-af50-50c5ef1b1202_010823-180713_88aba4772eef495ab6857a5076625ba9.tar.gz"
        )
        self.downloaded_file = Path("/tmp") / self.full_name.name

    @patch("catalystwan.session.ManagerSession")
    @patch("requests.Response")
    def test_delete_full_name(self, mock_session, mock_response):
        # Arrange
        mock_session.delete.return_value = mock_response
        mock_response.json.return_value = mock_response
        expected_api = f"/dataservice/tenantbackup/delete?fileName={self.full_name}"
        # Act
        TenantBackupRestoreAPI(mock_session).delete(str(self.full_name))
        # Assert
        mock_session.delete.assert_called_once_with(expected_api)

    @patch("catalystwan.session.ManagerSession")
    @patch("requests.Response")
    def test_delete_name(self, mock_session, mock_response):
        # Arrange
        mock_session.delete.return_value = mock_response
        mock_response.json.return_value = mock_response
        expected_api = f"/dataservice/tenantbackup/delete?fileName={self.full_name.name}"
        # Act
        TenantBackupRestoreAPI(mock_session).delete(str(self.full_name.name))
        # Assert
        mock_session.delete.assert_called_once_with(expected_api)

    @patch("catalystwan.session.ManagerSession")
    @patch("requests.Response")
    def test_delete_all(self, mock_session, mock_response):
        # Arrange
        mock_session.delete.return_value = mock_response
        mock_response.json.return_value = mock_response
        expected_api = "/dataservice/tenantbackup/delete?fileName=all"
        # Act
        TenantBackupRestoreAPI(mock_session).delete_all()
        # Assert
        mock_session.delete.assert_called_once_with(expected_api)

    @patch("catalystwan.session.ManagerSession")
    @patch("requests.Response")
    def test_download_full(self, mock_session, mock_response):
        # Arrange
        mock_session.get_file.return_value = mock_response
        mock_session.get.return_value = mock_response
        mock_response.status_code = 200
        mock_response.content = self.download_file_content
        with tempfile.TemporaryDirectory() as tmpdir:
            # Act
            download_path = TenantBackupRestoreAPI(mock_session).download(str(self.full_name), Path(tmpdir))
            # Assert
            self.assertEqual(download_path, Path(tmpdir) / self.full_name.name)

    @patch("catalystwan.session.ManagerSession")
    @patch("requests.Response")
    def test_download_name(self, mock_session, mock_response):
        # Arrange
        mock_session.get_file.return_value = mock_response
        mock_session.get.return_value = mock_response
        mock_response.status_code = 200
        mock_response.content = self.download_file_content
        with tempfile.TemporaryDirectory() as tmpdir:
            # Act
            download_path = TenantBackupRestoreAPI(mock_session).download(self.full_name.name, Path(tmpdir))
            # Assert
            self.assertEqual(download_path, Path(tmpdir) / self.full_name.name)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Final, List

from jinja2 import DebugUndefined, Environment, FileSystemLoader, meta  # type: ignore
from pydantic import BaseModel, ConfigDict, Field, field_validator

from catalystwan.utils.device_model import DeviceModel

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession

logger = logging.getLogger(__name__)


class GeneralTemplate(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    name: str = ""
    subTemplates: List[GeneralTemplate] = []

    templateId: str = ""
    templateType: str = ""


class DeviceTemplate(BaseModel):
    """
    ## Example:

    >>> templates = [
            "default_system", # Cisco System
            "default_logging", # Cisco Logging
            "default_banner", # Banner
        ]
    >>> device_template = DeviceTemplate(
            template_name="python",
            template_description="python",
            general_templates=templates
        )
    >>> session.api.templates.create(device_template)
    """

    template_name: str = Field(alias="templateName")
    template_description: str = Field(alias="templateDescription")
    general_templates: List[GeneralTemplate] = Field(alias="generalTemplates")
    device_role: str = Field(default="sdwan-edge", alias="deviceRole")
    device_type: DeviceModel = Field(alias="deviceType")
    security_policy_id: str = Field(default="", alias="securityPolicyId")
    policy_id: str = Field(default="", alias="policyId")

    def generate_payload(self) -> str:
        env = Environment(
            loader=FileSystemLoader(self.payload_path.parent),
            trim_blocks=True,
            lstrip_blocks=True,
            undefined=DebugUndefined,
        )
        template = env.get_template(self.payload_path.name)
        output = template.render(self.model_dump())

        ast = env.parse(output)
        if meta.find_undeclared_variables(ast):
            logger.info(meta.find_undeclared_variables(ast))
            raise Exception("There are undeclared variables.")
        return output

    @field_validator("general_templates", mode="before")
    @classmethod
    def parse_templates(cls, value):
        output = []
        for template in value:
            if isinstance(template, str):
                output.append(GeneralTemplate(name=template))
            else:
                output.append(template)
        return output

    payload_path: Final[Path] = Path(__file__).parent / "device_template_payload.json.j2"

    @classmethod
    def get(self, name: str, session: ManagerSession) -> DeviceTemplate:
        device_template = session.api.templates.get(DeviceTemplate).filter(name=name).single_or_default()
        resp = session.get(f"dataservice/template/device/object/{device_template.id}").json()
        return DeviceTemplate(**resp)

    model_config = ConfigDict(populate_by_name=True, use_enum_values=True)


class DeviceSpecificValue(BaseModel):
    property: str



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from enum import Enum
from pathlib import Path
from typing import ClassVar, List, Optional

from pydantic import ConfigDict, Field

from catalystwan.api.templates.bool_str import BoolStr
from catalystwan.api.templates.device_variable import DeviceVariable
from catalystwan.api.templates.feature_template import FeatureTemplate, FeatureTemplateValidator
from catalystwan.utils.timezone import Timezone


class MobileNumber(FeatureTemplateValidator):
    number: str


class SiteType(str, Enum):
    TYPE_1 = "type-1"
    TYPE_2 = "type-2"
    TYPE_3 = "type-3"
    CLOUD = "cloud"
    BRANCH = "branch"
    BR = "br"
    SPOKE = "spoke"


class ConsoleBaudRate(str, Enum):
    _1200 = "1200"
    _2400 = "2400"
    _4800 = "4800"
    _9600 = "9600"
    _19200 = "19200"
    _38400 = "38400"
    _57600 = "57600"
    _115200 = "115200"


class Protocol(str, Enum):
    TCP = "tcp"
    UDP = "udp"


class Boolean(str, Enum):
    OR = "or"
    AND = "and"


class Type(str, Enum):
    INTERFACE = "interface"
    STATIC_ROUTE = "static-route"


class Tracker(FeatureTemplateValidator):
    name: str
    endpoint_ip: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "endpoint-ip"})
    endpoint_ip_transport_port: Optional[str] = Field(
        default=None, json_schema_extra={"vmanage_key": "endpoint-ip", "data_path": ["endpoint-ip-transport-port"]}
    )
    protocol: Optional[Protocol] = Field(default=None, json_schema_extra={"data_path": ["endpoint-ip-transport-port"]})
    port: Optional[int] = Field(default=None, json_schema_extra={"data_path": ["endpoint-ip-transport-port"]})
    endpoint_dns_name: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "endpoint-dns-name"})
    endpoint_api_url: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "endpoint-api-url"})
    elements: Optional[List[str]] = None
    boolean: Optional[Boolean] = Boolean.OR
    threshold: Optional[int] = 300
    interval: Optional[int] = 60
    multiplier: Optional[int] = 3
    type: Optional[Type] = Type.INTERFACE
    model_config = ConfigDict(populate_by_name=True)


class Object(FeatureTemplateValidator):
    number: int


class ObjectTrack(FeatureTemplateValidator):
    object_number: int = Field(json_schema_extra={"vmanage_key": "object-number"})
    interface: str
    sig: str
    ip: str
    mask: Optional[str] = "0.0.0.0"
    vpn: int
    object: List[Object]
    boolean: Boolean
    model_config = ConfigDict(populate_by_name=True)


class Role(str, Enum):
    EDGE_ROUTER = "edge-router"
    BORDER_ROUTER = "border-router"


class AffinityPerVrf(FeatureTemplateValidator):
    affinity_group_number: Optional[int] = Field(
        default=None, json_schema_extra={"vmanage_key": "affinity-group-number"}
    )
    vrf_range: Optional[str] = Field(default=None, json_schema_extra={"vmanage_key": "vrf-range"})
    model_config = ConfigDict(populate_by_name=True)


class EnableMrfMigration(str, Enum):
    ENABLE = "enabled"
    ENABLE_FROM_BGP_CORE = "enabled-from-bgp-core"


class Vrf(FeatureTemplateValidator):
    vrf_id: int = Field(json_schema_extra={"vmanage_key": "vrf-id"})
    gateway_preference: Optional[List[int]] = Field(
        default=None, json_schema_extra={"vmanage_key": "gateway-preference"}
    )
    model_config = ConfigDict(populate_by_name=True)


class Epfr(str, Enum):
    DISABLED = "disabled"
    AGGRESSIVE = "aggressive"
    MODERATE = "moderate"
    CONSERVATIVE = "conservative"


class CiscoSystemModel(FeatureTemplate):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    timezone: Optional[Timezone] = Field(default=None, json_schema_extra={"data_path": ["clock"]})
    hostname: DeviceVariable = Field(
        default=DeviceVariable(name="system_host_name"),
        validate_default=True,
        json_schema_extra={"vmanage_key": "host-name"},
    )
    location: Optional[str] = None
    latitude: Optional[float] = Field(default=None, json_schema_extra={"data_path": ["gps-location"]})
    longitude: Optional[float] = Field(default=None, json_schema_extra={"data_path": ["gps-location"]})
    range: Optional[int] = Field(100, json_schema_extra={"data_path": ["gps-location", "geo-fencing"]})
    enable_fencing: Optional[BoolStr] = Field(
        default=None, json_schema_extra={"data_path": ["gps-location", "geo-fencing"], "vmanage_key": "enable"}
    )
    mobile_number: Optional[List[MobileNumber]] = Field(
        default=None,
        json_schema_extra={"vmanage_key": "mobile-number", "data_path": ["gps-location", "geo-fencing", "sms"]},
    )
    enable_sms: Optional[BoolStr] = Field(
        default=False, json_schema_extra={"data_path": ["gps-location", "geo-fencing", "sms"], "vmanage_key": "enable"}
    )
    device_groups: Optional[List[str]] = Field(default=None, json_schema_extra={"vmanage_key": "device-groups"})
    controller_group_list: Optional[List[int]] = Field(
        default=None, json_schema_extra={"vmanage_key": "controller-group-list"}
    )
    system_ip: DeviceVariable = Field(
        default=DeviceVariable(name="system_system_ip"), json_schema_extra={"vmanage_key": "system-ip"}
    )
    overlay_id: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "overlay-id"})
    site_id: int = Field(default=DeviceVariable(name="system_site_id"), json_schema_extra={"vmanage_key": "site-id"})
    site_type: Optional[List[SiteType]] = Field(default=None, json_schema_extra={"vmanage_key": "site-type"})
    port_offset: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "port-offset"})
    port_hop: Optional[BoolStr] = Field(default=None, json_schema_extra={"vmanage_key": "port-hop"})
    control_session_pps: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "control-session-pps"})
    track_transport: Optional[BoolStr] = Field(default=None, json_schema_extra={"vmanage_key": "track-transport"})
    track_interface_tag: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "track-interface-tag"})
    console_baud_rate: Optional[ConsoleBaudRate] = Field(
        default=None, json_schema_extra={"vmanage_key": "console-baud-rate"}
    )
    max_omp_sessions: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "max-omp-sessions"})
    multi_tenant: Optional[BoolStr] = Field(default=None, json_schema_extra={"vmanage_key": "multi-tenant"})
    track_default_gateway: Optional[BoolStr] = Field(
        default=None, json_schema_extra={"vmanage_key": "track-default-gateway"}
    )
    admin_tech_on_failure: Optional[BoolStr] = Field(
        default=None, json_schema_extra={"vmanage_key": "admin-tech-on-failure"}
    )
    enable_tunnel: Optional[BoolStr] = Field(
        default=None, json_schema_extra={"vmanage_key": "enable", "data_path": ["on-demand"]}
    )
    idle_timeout: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "idle-timeout"})
    on_demand_idle_timeout_min: Optional[int] = Field(
        default=None, json_schema_extra={"vmanage_key": "idle-timeout", "data_path": ["on-demand"]}
    )
    tracker: Optional[List[Tracker]] = None
    object_track: Optional[List[ObjectTrack]] = Field(default=None, json_schema_extra={"vmanage_key": "object-track"})
    region_id: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "region-id"})
    secondary_region: Optional[int] = Field(default=None, json_schema_extra={"vmanage_key": "secondary-region"})
    role: Optional[Role] = None
    affinity_group_number: Optional[int] = Field(
        default=None, json_schema_extra={"vmanage_key": "affinity-group-number", "data_path": ["affinity-group"]}
    )
    preference: Optional[List[int]] = Field(default=None, json_schema_extra={"data_path": ["affinity-group"]})
    preference_auto: Optional[BoolStr] = Field(default=None, json_schema_extra={"vmanage_key": "preference-auto"})
    affinity_per_vrf: Optional[List[AffinityPerVrf]] = Field(
        default=None, json_schema_extra={"vmanage_key": "affinity-per-vrf"}
    )
    transport_gateway: Optional[BoolStr] = Field(default=None, json_schema_extra={"vmanage_key": "transport-gateway"})
    enable_mrf_migration: Optional[EnableMrfMigration] = Field(
        default=None, json_schema_extra={"vmanage_key": "enable-mrf-migration"}
    )
    migration_bgp_community: Optional[int] = Field(
        default=None, json_schema_extra={"vmanage_key": "migration-bgp-community"}
    )
    enable_management_region: Optional[BoolStr] = Field(
        default=None, json_schema_extra={"vmanage_key": "enable-management-region"}
    )
    vrf: Optional[List[Vrf]] = None
    management_gateway: Optional[BoolStr] = Field(default=None, json_schema_extra={"vmanage_key": "management-gateway"})
    epfr: Optional[Epfr] = None

    payload_path: ClassVar[Path] = Path(__file__).parent / "DEPRECATED"
    type: ClassVar[str] = "cisco_system"



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from pydantic import AliasPath, ConfigDict, Field, field_validator

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global


class ExpandedCommunityParcel(_ParcelBase):
    model_config = ConfigDict(populate_by_name=True)
    expandedCommunityList: Global[list] = Field(
        default=as_global([]),
        serialization_alias="expandedCommunityList",
        validation_alias=AliasPath("data", "expandedCommunityList"),
    )

    def add_community(self, expanded_community: str):
        self.expandedCommunityList.value.append(expanded_community)

    @field_validator("expandedCommunityList")
    @classmethod
    def check_rate(cls, expanded_community_list: Global):
        assert all([isinstance(ec, str) for ec in expanded_community_list.value])
        return expanded_community_list



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable
from catalystwan.models.common import ServiceChainNumber

Action = Literal[
    "drop",
    "accept",
]

IcmpMessage = Literal[
    "administratively-prohibited",
    "dod-host-prohibited",
    "dod-net-prohibited",
    "echo",
    "echo-reply",
    "echo-reply-no-error",
    "extended-echo",
    "extended-echo-reply",
    "general-parameter-problem",
    "host-isolated",
    "host-precedence-unreachable",
    "host-redirect",
    "host-tos-redirect",
    "host-tos-unreachable",
    "host-unknown",
    "host-unreachable",
    "interface-error",
    "malformed-query",
    "multiple-interface-match",
    "net-redirect",
    "net-tos-redirect",
    "net-tos-unreachable",
    "net-unreachable",
    "network-unknown",
    "no-room-for-option",
    "option-missing",
    "packet-too-big",
    "parameter-problem",
    "photuris",
    "port-unreachable",
    "precedence-unreachable",
    "protocol-unreachable",
    "reassembly-timeout",
    "redirect",
    "router-advertisement",
    "router-solicitation",
    "source-route-failed",
    "table-entry-error",
    "time-exceeded",
    "timestamp-reply",
    "timestamp-request",
    "ttl-exceeded",
    "unreachable",
]

Icmp6Message = Literal[
    "beyond-scope",
    "cp-advertisement",
    "cp-solicitation",
    "destination-unreachable",
    "dhaad-reply",
    "dhaad-request",
    "echo-reply",
    "echo-request",
    "header",
    "hop-limit",
    "ind-advertisement",
    "ind-solicitation",
    "mld-query",
    "mld-reduction",
    "mld-report",
    "mldv2-report",
    "mpd-advertisement",
    "mpd-solicitation",
    "mr-advertisement",
    "mr-solicitation",
    "mr-termination",
    "nd-na",
    "nd-ns",
    "next-header-type",
    "ni-query",
    "ni-query-name",
    "ni-query-v4-address",
    "ni-query-v6-address",
    "ni-response",
    "ni-response-qtype-unknown",
    "ni-response-refuse",
    "ni-response-success",
    "no-admin",
    "no-route",
    "packet-too-big",
    "parameter-option",
    "parameter-problem",
    "port-unreachable",
    "reassembly-timeout",
    "redirect",
    "reject-route",
    "renum-command",
    "renum-result",
    "renum-seq-number",
    "router-advertisement",
    "router-renumbering",
    "router-solicitation",
    "rpl-control",
    "source-policy",
    "source-route-header",
    "time-exceeded",
    "unreachable",
]


class SourceDataIPv4Prefix(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    source_ip_prefix: Union[Global[str], Variable] = Field(
        serialization_alias="sourceIpPrefix", validation_alias="sourceIpPrefix"
    )


class SourceDataIPv6Prefix(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    source_ip_prefix: Union[Global[str], Variable] = Field(
        serialization_alias="sourceIpPrefix", validation_alias="sourceIpPrefix"
    )


class SourceDataIPv4PrefixParcel(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    source_data_prefix_list: Global[UUID] = Field(
        serialization_alias="sourceDataPrefixList", validation_alias="sourceDataPrefixList"
    )


class SourceDataIPv6PrefixParcel(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    source_data_prefix_list: Global[UUID] = Field(
        serialization_alias="sourceDataPrefixList", validation_alias="sourceDataPrefixList"
    )


class SourcePort(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    source_port: Global[int] = Field(serialization_alias="sourcePort", validation_alias="sourcePort")


class DestinationDataIPv4Prefix(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    destination_ip_prefix: Union[Global[str], Variable] = Field(
        serialization_alias="destinationIpPrefix", validation_alias="destinationIpPrefix"
    )


class DestinationDataIPv6Prefix(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    destination_ip_prefix: Union[Global[str], Variable] = Field(
        serialization_alias="destinationIpPrefix", validation_alias="destinationIpPrefix"
    )


class DestinationDataIPv4PrefixParcel(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    destination_data_prefix_list: Global[UUID] = Field(
        serialization_alias="destinationDataPrefixList", validation_alias="destinationDataPrefixList"
    )


class DestinationDataIPv6PrefixParcel(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    destination_data_prefix_list: Global[UUID] = Field(
        serialization_alias="destinationDataPrefixList", validation_alias="destinationDataPrefixList"
    )


class DestinationPort(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    destination_port: Global[int] = Field(serialization_alias="destinationPort", validation_alias="destinationPort")


TcpState = Literal["syn"]


class IPv4Match(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    dscp: Optional[Global[List[int]]] = None
    packet_length: Optional[Global[int]] = Field(
        serialization_alias="packetLength", validation_alias="packetLength", default=None
    )
    protocol: Optional[Global[List[int]]] = None
    icmp_message: Optional[Global[List[IcmpMessage]]] = Field(
        serialization_alias="icmpMsg", validation_alias="icmpMsg", default=None
    )
    source_data_prefix: Optional[Union[SourceDataIPv4Prefix, SourceDataIPv4PrefixParcel]] = Field(
        serialization_alias="sourceDataPrefix", validation_alias="sourceDataPrefix", default=None
    )
    source_ports: Optional[List[SourcePort]] = Field(
        serialization_alias="sourcePorts", validation_alias="sourcePorts", default=None
    )
    destination_data_prefix: Optional[Union[DestinationDataIPv4Prefix, DestinationDataIPv4PrefixParcel]] = Field(
        serialization_alias="destinationDataPrefix", validation_alias="destinationDataPrefix", default=None
    )
    destination_ports: Optional[List[DestinationPort]] = Field(
        serialization_alias="destinationPorts", validation_alias="destinationPorts", default=None
    )
    tcp: Optional[Global[TcpState]] = None


class IPv6Match(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    next_header: Optional[Global[int]] = Field(
        serialization_alias="nextHeader", validation_alias="nextHeader", default=None
    )
    packet_length: Optional[Global[int]] = Field(
        serialization_alias="packetLength", validation_alias="packetLength", default=None
    )
    source_data_prefix: Optional[Union[SourceDataIPv6Prefix, SourceDataIPv6PrefixParcel]] = Field(
        serialization_alias="sourceDataPrefix", validation_alias="sourceDataPrefix", default=None
    )
    source_ports: Optional[List[SourcePort]] = Field(
        serialization_alias="sourcePorts", validation_alias="sourcePorts", default=None
    )
    destination_data_prefix: Optional[Union[DestinationDataIPv6Prefix, DestinationDataIPv6PrefixParcel]] = Field(
        serialization_alias="destinationDataPrefix", validation_alias="destinationDataPrefix", default=None
    )
    destination_ports: Optional[List[DestinationPort]] = Field(
        serialization_alias="destinationPorts", validation_alias="destinationPorts", default=None
    )
    tcp: Optional[Global[TcpState]] = None
    traffic_class: Optional[Global[int]] = None
    icmp6_message: Optional[Global[List[Icmp6Message]]] = Field(
        serialization_alias="icmpMsg", validation_alias="icmpMsg", default=None
    )


class ServiceChain(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    service_chain_number: Union[Global[ServiceChainNumber], Variable] = Field(
        serialization_alias="serviceChainNumber", validation_alias="serviceChainNumber"
    )
    vpn: Optional[Union[Global[int], Variable]] = None
    fallback: Optional[Union[Global[bool], Variable, Default[bool]]] = None


class AcceptActionIPv4(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    set_dscp: Optional[Global[int]] = Field(serialization_alias="setDscp", validation_alias="setDscp", default=None)
    counter_name: Optional[Global[str]] = Field(
        serialization_alias="counterName", validation_alias="counterName", default=None
    )
    log: Optional[Union[Global[bool], Default[bool]]] = None
    set_next_hop: Optional[Global[str]] = Field(
        serialization_alias="setNextHop", validation_alias="setNextHop", default=None
    )
    set_service_chain: Optional[ServiceChain] = Field(
        serialization_alias="setServiceChain", validation_alias="setServiceChain", default=None
    )
    mirror: Optional[Global[UUID]] = None
    policer: Optional[Global[UUID]] = None


class AcceptActionIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    counter_name: Optional[Global[str]] = Field(
        serialization_alias="counterName", validation_alias="counterName", default=None
    )
    log: Optional[Union[Global[bool], Default[bool]]] = None
    set_next_hop: Optional[Global[str]] = Field(
        serialization_alias="setNextHop", validation_alias="setNextHop", default=None
    )
    set_service_chain: Optional[ServiceChain] = Field(
        serialization_alias="setServiceChain", validation_alias="setServiceChain", default=None
    )
    set_traffic_class: Optional[Global[int]] = Field(
        serialization_alias="setTrafficClass", validation_alias="setTrafficClass", default=None
    )
    mirror: Optional[Global[UUID]] = None
    policer: Optional[Global[UUID]] = None


class DropAction(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    counter_name: Optional[Global[str]] = Field(
        serialization_alias="counterName", validation_alias="counterName", default=None
    )
    log: Optional[Union[Global[bool], Default[bool]]] = None


class AcceptActionsIPv4(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    accept: AcceptActionIPv4


class AcceptActionsIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    accept: AcceptActionIPv6


class DropActions(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    drop: DropAction


class IPv4SequenceBaseAction(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    sequence_id: Global[int] = Field(serialization_alias="sequenceId", validation_alias="sequenceId")
    sequence_name: Global[str] = Field(serialization_alias="sequenceName", validation_alias="sequenceName")
    base_action: Union[Global[Action], Default[Action]] = Field(
        serialization_alias="baseAction", validation_alias="baseAction", default=Default[Action](value="accept")
    )
    match_entries: Optional[List[IPv4Match]] = Field(
        serialization_alias="matchEntries", validation_alias="matchEntries", default=None
    )


class IPv6SequenceBaseAction(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    sequence_id: Global[int] = Field(serialization_alias="sequenceId", validation_alias="sequenceId")
    sequence_name: Global[str] = Field(serialization_alias="sequenceName", validation_alias="sequenceName")
    base_action: Union[Global[Action], Default[Action]] = Field(
        serialization_alias="baseAction", validation_alias="baseAction", default=Default[Action](value="accept")
    )
    match_entries: Optional[List[IPv6Match]] = Field(
        serialization_alias="matchEntries", validation_alias="matchEntries", default=None
    )


class IPv4SequenceActions(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    sequence_id: Global[int] = Field(serialization_alias="sequenceId", validation_alias="sequenceId")
    sequence_name: Global[str] = Field(serialization_alias="sequenceName", validation_alias="sequenceName")
    actions: List[Union[AcceptActionsIPv4, DropActions]] = Field(
        serialization_alias="actions", validation_alias="actions"
    )
    match_entries: Optional[List[IPv4Match]] = Field(
        serialization_alias="matchEntries", validation_alias="matchEntries", default=None
    )


class IPv6SequenceActions(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    sequence_id: Global[int] = Field(serialization_alias="sequenceId", validation_alias="sequenceId")
    sequence_name: Global[str] = Field(serialization_alias="sequenceName", validation_alias="sequenceName")
    actions: List[Union[AcceptActionsIPv6, DropActions]] = Field(
        serialization_alias="actions", validation_alias="actions"
    )
    match_entries: Optional[List[IPv6Match]] = Field(
        serialization_alias="matchEntries", validation_alias="matchEntries", default=None
    )


class IPv4AclData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    defautl_action: Union[Global[Action], Default[Action]] = Field(
        serialization_alias="defaultAction", validation_alias="defaultAction", default=Default[Action](value="drop")
    )
    sequences: List[Union[IPv4SequenceBaseAction, IPv4SequenceActions]]


class IPv4AclCreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: IPv4AclData


class IPv6AclData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    defautl_action: Union[Global[Action], Default[Action]] = Field(
        serialization_alias="defaultAction", validation_alias="defaultAction", default=Default[Action](value="drop")
    )
    sequences: List[Union[IPv6SequenceBaseAction, IPv6SequenceActions]]


class IPv6AclCreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: IPv6AclData



# Copyright 2024 Cisco Systems, Inc. and its affiliates

#  type: ignore
from catalystwan.tests.templates.models.cisco_aaa import cisco_aaa, complex_aaa_model
from catalystwan.tests.templates.models.cisco_banner import banner_model
from catalystwan.tests.templates.models.cisco_bfd import bfd_model
from catalystwan.tests.templates.models.cisco_vpn import basic_cisco_vpn, complex_vpn_model
from catalystwan.tests.templates.models.omp_vsmart import default_omp, omp_2, omp_3

__all__ = [
    "default_omp",
    "omp_2",
    "omp_3",
    "cisco_aaa",
    "complex_aaa_model",
    "basic_cisco_vpn",
    "complex_vpn_model",
    "banner_model",
    "bfd_model",
]



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from uuid import UUID

from catalystwan.endpoints import APIEndpoints, delete, get, post, put
from catalystwan.models.policy.lists import PrefixList
from catalystwan.models.policy.policy_list import (
    InfoTag,
    PolicyListEndpoints,
    PolicyListId,
    PolicyListInfo,
    PolicyListPreview,
)
from catalystwan.typed_list import DataSequence


class PrefixListEditPayload(PrefixList, PolicyListId):
    pass


class PrefixListInfo(PrefixList, PolicyListInfo):
    pass


class ConfigurationPolicyPrefixList(APIEndpoints, PolicyListEndpoints):
    @post("/template/policy/list/prefix")
    def create_policy_list(self, payload: PrefixList) -> PolicyListId:
        ...

    @delete("/template/policy/list/prefix/{id}")
    def delete_policy_list(self, id: UUID) -> None:
        ...

    @delete("/template/policy/list/prefix")
    def delete_policy_lists_with_info_tag(self, params: InfoTag) -> None:
        ...

    @put("/template/policy/list/prefix/{id}")
    def edit_policy_list(self, id: UUID, payload: PrefixListEditPayload) -> None:
        ...

    @get("/template/policy/list/prefix/{id}")
    def get_lists_by_id(self, id: UUID) -> PrefixListInfo:
        ...

    @get("/template/policy/list/prefix", "data")
    def get_policy_lists(self) -> DataSequence[PrefixListInfo]:
        ...

    @get("/template/policy/list/prefix/filtered", "data")
    def get_policy_lists_with_info_tag(self, params: InfoTag) -> DataSequence[PrefixListInfo]:
        ...

    @post("/template/policy/list/prefix/preview")
    def preview_policy_list(self, payload: PrefixList) -> PolicyListPreview:
        ...

    @get("/template/policy/list/prefix/preview/{id}")
    def preview_policy_list_by_id(self, id: UUID) -> PolicyListPreview:
        ...
